import asyncio
import contextvars
import time
from contextlib import ExitStack, asynccontextmanager
from dataclasses import dataclass, field
from typing import IO, TYPE_CHECKING
from dagon.ui import ansi

from dagon.ui.events import UIEvents
from dagon.ui.message import Message, MessageType
from dagon.ui.proc import PrintProcessResultUIInfo, make_proc_info_box

from .. import util
from ..ext.iface import OpaqueTaskGraphView
from ..task.dag import OpaqueTask
from .iface import I_UIExtension
from .stdio_cap import CapturedOutput, OutputAccumulator, capture_std_output


@dataclass()
class _StdioAccum:
    capture: CapturedOutput
    stdout_acc: OutputAccumulator = field(default_factory=OutputAccumulator)
    stderr_acc: OutputAccumulator = field(default_factory=OutputAccumulator)


_GLBL_CTX = contextvars.ContextVar[_StdioAccum]('_CAPTURE_CTX')

_TASK_CTX = contextvars.ContextVar[OpaqueTask]('_TASK_CTX')


class SimpleUI:
    dagon_ui_name = 'dagon.ui.simple'
    dagon_ui_opt_name = 'simple'

    @asynccontextmanager
    async def ui_global_context(self, graph: OpaqueTaskGraphView, events: UIEvents):
        with ExitStack() as st:
            cap = st.enter_context(capture_std_output())
            st.enter_context(util.scope_set_contextvar(_GLBL_CTX, _StdioAccum(cap)))
            start = time.time()
            print(f'[dagon]: {len(tuple(graph.all_nodes))} tasks to run')
            loop = asyncio.get_event_loop()
            st.enter_context(cap.on_out.connect(lambda s: loop.call_soon_threadsafe(lambda: self._append_stdout(s))))
            st.enter_context(cap.on_err.connect(lambda s: loop.call_soon_threadsafe(lambda: self._append_stderr(s))))
            st.enter_context(events.message.connect(self._on_message))
            st.enter_context(events.status.connect(self._on_status))
            st.enter_context(events.print_process_done.connect(self._echo_proc))
            yield
            print(f'[dagon] Finished in {time.time() - start:.4}s')
            return

    @asynccontextmanager
    async def ui_task_context(self, task: OpaqueTask):
        print(f'[begin:{task.name}]')
        with util.scope_set_contextvar(_TASK_CTX, task):
            yield
        print(f'[end:{task.name}]')

    def _on_message(self, msg: Message) -> None:
        if msg.type != MessageType.Print:
            self._append_stderr(msg.content + '\n')
        else:
            self._append_stdout(msg.content + '\n')

    def _on_status(self, status: str) -> None:
        self._append_stdout(status)

    def _echo_proc(self, result: PrintProcessResultUIInfo) -> None:
        for m in make_proc_info_box(result, max_width=ansi.get_term_width()):
            self._on_message(m)

    def _append_stdout(self, s: str) -> None:
        ctx = _GLBL_CTX.get()
        self._append(s, ctx.capture.real_stdout, ctx.stdout_acc)

    def _append_stderr(self, s: str) -> None:
        ctx = _GLBL_CTX.get()
        self._append(s, ctx.capture.real_stderr, ctx.stderr_acc)

    def _append(self, s: str, into: IO[str], acc: OutputAccumulator) -> None:
        acc.append(s)
        for l in acc.take_lines():
            task = _TASK_CTX.get(None)
            if task is not None:
                prefix = f'[dagon task:{task.name}] '.encode(errors='?')
            else:
                prefix = b'[dagon]'
            l = prefix + l
            into.write(l.decode(encoding='utf-8', errors='?'))
        into.flush()


if TYPE_CHECKING:
    util.typecheck(I_UIExtension)(SimpleUI)



import itertools
from contextlib import ExitStack
from typing import Any, Callable, Coroutine

import dagon.task as mod
from dagon.ext.exec import ExtAwareExecutor
from dagon.tool import main
from dagon.util.testing import async_test

from ..core.result import Failure, NodeResult, Success
from .dag import TaskDAG, result_of

# pyright: reportUnusedFunction=false


def test_add() -> None:
    dag = TaskDAG('test')

    @mod.define_in(dag)
    async def meow() -> None:
        pass

    @mod.define_in(dag, depends=[meow])
    async def inner() -> int:
        return 0


@async_test
async def test_add_and_run() -> None:
    dag = TaskDAG('test')

    @mod.define_in(dag)
    async def string() -> str:
        return 'hello'

    @mod.define_in(dag, depends=[string])
    async def print_string() -> None:
        await result_of(string)

    @mod.define_in(dag, depends=[string])
    async def print_string2() -> None:
        await result_of(string)
        await result_of(string)

    @mod.define_in(dag, depends=[print_string, print_string2])
    async def final_tgt() -> None:
        pass

    results = await dag.execute(['print-string'])
    assert set(results.values()) == {
        NodeResult(string, Success('hello')),
        NodeResult(print_string, Success(None)),
    }

    results = await dag.execute(['print-string2'])
    assert set(results.values()) == {
        NodeResult(string, Success('hello')),
        NodeResult(print_string2, Success(None)),
    }

    results = await dag.execute(['final-tgt'])
    assert set(results.values()) == {
        NodeResult(string, Success('hello')),
        NodeResult(print_string, Success(None)),
        NodeResult(print_string2, Success(None)),
        NodeResult(final_tgt, Success(None)),
    }


async def _run_oo_test(use_oo_deps: bool) -> None:
    value = 0

    dag = TaskDAG('test')

    @mod.define_in(dag)
    async def _first() -> None:
        nonlocal value
        value = 1

    @mod.define_in(dag, order_only_dependss=[_first])
    async def _second() -> None:
        nonlocal value
        if use_oo_deps:
            assert value == 1, 'Value should have been set'
        else:
            assert value == 0, 'Value should not have been set'

    if use_oo_deps:
        result = await dag.execute([_first.name, _second.name])
    else:
        result = await dag.execute([_second.name])

    if use_oo_deps:
        assert set(result.values()) == {
            NodeResult(_first, Success(None)),
            NodeResult(_second, Success(None)),
        }
    else:
        assert set(result.values()) == {NodeResult(_second, Success(None))}


@async_test
async def test_order_only() -> None:
    for _, use_oo_deps in itertools.product(range(100), (True, False)):
        # Run this repeatedly to try and force the ordering to fail
        await _run_oo_test(use_oo_deps)


def run_test_on_fun(fn: Callable[[], Coroutine[None, None, None]], **kw: Any) -> None:
    with ExitStack() as st:
        exts = main.get_extensions()
        st.enter_context(exts.app_context())
        t = mod.create_task_from_function(fn, **kw)
        dag = TaskDAG('Test')
        dag.add_task(t)
        graph = dag.low_level_graph([t.name])
        results = ExtAwareExecutor(exts, graph, catch_signals=False, fail_cancels=True).run_all_until_complete()
        for f in results.values():
            if isinstance(f.result, Failure):
                f.result.reraise()
            assert isinstance(f.result, Success)



import enum

import pytest

from . import ext as mod


class MyEnum(enum.Enum):
    Value = 'value'
    Cat = 'cat'


def test_option_set() -> None:
    s = mod.OptionSet()
    opt = mod.Option('key', type=int, parse=int)
    s.add(opt)
    assert s.get('key') is opt


def test_fulfill() -> None:
    s = mod.OptionSet()
    name = s.add(mod.Option('name', type=str, parse=str))
    age = s.add(
        mod.Option('age',
                   type=int,
                   parse=int,
                   default=None,
                   validate=lambda i: 'Must be a positive number' if i and i < 0 else None))
    # Invalid option name:
    with pytest.raises(NameError):
        s.fulfill(('invalid=12', ))
    # Invalid option string type:
    with pytest.raises(ValueError):
        s.fulfill(['age=cats'])
    # Validation failure:
    with pytest.raises(RuntimeError):
        s.fulfill(['age=-3'])
    # Good fulfillment:
    ff = s.fulfill(('name=John Doe', 'age=11'))
    assert ff.get(name) == 'John Doe'
    assert ff.get(age) == 11
    # Fulfill with missing option:
    ff = s.fulfill(['name=Jane Smith'])
    assert ff.get(name) == 'Jane Smith'
    assert ff.get(age) is None

    # Add a default value
    loc = s.add(mod.Option('loc', type=str, parse=str, default='The moon'))
    ff = s.fulfill(['name=Joe=equals'])
    assert ff.get(name) == 'Joe=equals'
    assert ff.get(loc) == 'The moon'

    # Get option value by name
    assert ff.get('name') == 'Joe=equals'

    # Enum options
    some_value = s.add(mod.Option('something', type=MyEnum, parse=MyEnum))
    with pytest.raises(ValueError):
        s.fulfill(['something=Dogs'])
    ff = s.fulfill(['something=cat'])
    assert ff.get(some_value) == MyEnum.Cat



import argparse
from typing import Sequence

from dagon.ext.loader import ExtLoader
from typing_extensions import Protocol


class ParsedArgs(Protocol):
    tasks: Sequence[str]
    list_tasks: bool
    no_doc: bool
    fail_cancels: bool
    debug: bool


def get_argparser(exts: ExtLoader) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument('tasks', nargs='*', help='Tasks to execute')
    parser.add_argument('-lt', '--list-tasks', action='store_true')
    parser.add_argument('--no-doc', action='store_true', help='Do not print docstrings when listing tasks/options')
    parser.add_argument('-j', '--jobs', type=int, help='Maximum number of parallel jobs to run')
    parser.add_argument('--fail-cancels',
                        action='store_true',
                        help='Cancel running tasks immediately when a task fails')
    parser.add_argument('--debug', action='store_true', help='Enable debug output')
    exts.add_options(parser)
    return parser



from __future__ import annotations

from typing import NamedTuple
from dagon.core.result import NodeResult

from dagon.task.dag import OpaqueTask
from dagon.ui.message import Message
from dagon.ui.proc import PrintProcessResultUIInfo

from ..event import Event


class ProgressInfo(NamedTuple):
    progress: float | None


class UIEvents(NamedTuple):
    message: Event[Message] = Event()
    status: Event[str] = Event()
    task_result: Event[NodeResult[OpaqueTask]] = Event()
    progress: Event[ProgressInfo] = Event()
    print_process_done: Event[PrintProcessResultUIInfo] = Event()



from .exec import SimpleExecutor
from .ll_dag import LowLevelDAG
from .result import Cancellation, Failure, NodeResult, Success

__all__ = [
    'LowLevelDAG',
    'NodeResult',
    'Success',
    'Failure',
    'Cancellation',
    'SimpleExecutor',
]



"""
Support for ANSI control sequences
"""
from __future__ import annotations
import re
from typing import cast

import os
import sys

ESC = '\x1b'
CSI = ESC + '['
NORMAL = CSI + '0m'
INVERSE = CSI + '7m'
NO_INVERSE = CSI + '27m'


def is_ansi_supported(fileno: int | None = None) -> bool:
    if fileno is None:
        fileno = sys.__stdout__.fileno()

    if not os.isatty(fileno):
        return False

    if os.name != 'nt':
        # Nix platforms support ANSI on TTYs by default
        return True

    import msvcrt  # pylint: disable=import-error
    from ctypes import byref, windll  # type: ignore
    from ctypes.wintypes import DWORD
    try:
        handle = msvcrt.get_osfhandle(fileno)  # type: ignore
    except OSError as e:
        if e.errno != 9:
            raise
        # Failed to get a HANDLE for the given fileno
        return False

    mode = DWORD()
    ok = windll.kernel32.GetConsoleMode(handle, byref(mode))
    assert ok, windll.kernel32.GetLastError()
    return (mode.value & 0x4) != 0


def ensure_ansi_term(fileno: int | None = None) -> bool:
    if not fileno:
        fileno = sys.__stdout__.fileno()

    if os.name != 'nt':
        # We only need to do this work for WIndows
        return is_ansi_supported(fileno)

    import msvcrt  # pylint: disable=import-error
    from ctypes import byref, windll  # type: ignore
    from ctypes.wintypes import DWORD
    try:
        handle = msvcrt.get_osfhandle(fileno)  # type: ignore
    except OSError as e:
        if e.errno != 9:
            raise
        # Failed to get a HANDLE for the given fileno
        return False

    mode = DWORD()
    ok = windll.kernel32.GetConsoleMode(handle, byref(mode))
    assert ok, windll.kernel32.GetLastError()
    new_mode = int(mode.value) | 0x4
    ok = windll.kernel32.SetConsoleMode(handle, new_mode)
    return bool(ok)


def style(style_code: str, string: str, force: bool = False) -> str:
    if not force and not os.isatty(sys.__stdout__.fileno()):
        return string
    return f'{CSI}{style_code}m{string}{NORMAL}'


def bold_green(s: str) -> str:
    return style('1;32', s)


def bold_red(s: str) -> str:
    return style('1;31', s)


def bold_cyan(s: str) -> str:
    return style('1;36', s)


def bold_yellow(s: str) -> str:
    return style('1;33', s)


def bold(s: str) -> str:
    return style('1', s)


def get_term_width() -> int:
    import struct
    if os.name == 'nt':
        # Windows has a special way to get the console dims
        from ctypes import create_string_buffer, windll  # type: ignore
        handle = windll.kernel32.GetStdHandle(-11)  # -11 == STD_OUTPUT_HANDLE
        buf = create_string_buffer(22)
        okay = windll.kernel32.GetConsoleScreenBufferInfo(handle, buf)
        if not okay:
            return 80
        info = struct.unpack('hhhhHhhhhhh', buf.raw)
        return cast(int, info[7] - info[5])

    import termios
    from fcntl import ioctl
    buf2 = ioctl(sys.__stdout__, termios.TIOCGWINSZ, b'\000' * 8)
    width: int = struct.unpack('hhhh', buf2)[1]
    return width


ESCAPE_RE = re.compile(r'\x1b(\[|\()[0-?]*[ -/]*[@-~]')


def strip_escapes(s: str) -> str:
    return ESCAPE_RE.sub('', s)



"""
Module ``dagon.core.ll_dag``
############################

This module provides a low-level generic task-aware DAG implementation.

The main class is `.LowLevelDAG`, which tracks node states and
dependencies between nodes.

No semantics are imposed on the node objects given to a `.LowLevelDAG`,
other than they must be valid as keys in a `set` or `dict` (They must be
`Hashable`).

This class neither concerns itself with "executing" a node. It is up to a user
of `.LowLevelDAG` to execute the nodes.

A simple low-level execution API can be found in `dagon.core.exec`.
"""
from __future__ import annotations

import copy
import enum
from dataclasses import dataclass, field
from typing import Any, Dict, Generic, Hashable, Iterable, Sequence, TypeVar

from typing_extensions import TypeAlias

from dagon.util import fixup_dataclass_docs

NodeT = TypeVar('NodeT', bound=Hashable)
"""
A type parameter for directed graph nodes. Each node must be hashable.
"""


class GraphError(RuntimeError):
    """
    Base class for any errors involving the directed graph
    """


@fixup_dataclass_docs
@dataclass()
class GraphStateError(RuntimeError):
    """
    Raised by an attempted operation on the graph when the graph is in an
    invalid state for that operation.
    """
    node: Any
    "The node that is in an invalid state"


@fixup_dataclass_docs
@dataclass
class CycleError(GraphError, Generic[NodeT]):
    """
    Exception raised when adding an edge to a graph would create a dependency cycle.
    """
    hint: Sequence[NodeT]
    """The sequence of nodes that were discovered forming a cycle"""


@fixup_dataclass_docs
@dataclass()
class DuplicateNodeError(GraphError):
    """
    Exception raised when attempting to add a node that already exists within the graph.
    """
    node: Any
    """The node that is already present in the graph"""


@fixup_dataclass_docs
@dataclass()
class DuplicateEdgeError(GraphError):
    """
    Exception raised when attempting to add a duplicate edge to a graph.
    """
    edge: Edge[Any]
    """The edge that was attempted to be added"""


@fixup_dataclass_docs
@dataclass()
class MissingNodeError(GraphError):
    """
    Exception raised when requesting data for a node that is not a member of the graph.
    """
    node: Any
    """The node that was requested"""


class NodeState(enum.Enum):
    """The task-state of a node within the graph"""
    Pending = 0
    "The node has not been marked running nor marked finished"
    Running = 1
    "The node has been marked as running but has not been marked finished"
    Finished = 2
    "The node has been marked as finished"


@fixup_dataclass_docs
@dataclass()
class _GraphNodeMeta(Generic[NodeT]):
    """
    A node in the graph, containing an associated value.
    """
    value: NodeT
    'The object associated with this node'
    pending_inputs: int = 0
    'The number of pending inputs to this node in the current execution'
    order: int = 0
    'The order/"depth" of the node in the graph'
    state: NodeState = NodeState.Pending
    'The current state of the graph'
    outs: set[NodeT] = field(default_factory=set)
    'The objects that depends on this node in the graph'
    ins: set[NodeT] = field(default_factory=set)
    'The objects upon which this node depends'

    def __deepcopy__(self, memo: Any) -> _GraphNodeMeta[NodeT]:
        return _GraphNodeMeta(self.value, self.pending_inputs, self.order, self.state, set(self.outs), set(self.ins))

    @property
    def is_ready(self) -> bool:
        """Determine whether this node is ready to run"""
        return self.state is NodeState.Pending and self.pending_inputs == 0


@dataclass(frozen=True)
class Edge(Generic[NodeT]):
    """
    An edge in the graph representing a dependency of one node value on another.
    """
    from_: NodeT
    "The outgoing node in the edge. This is the dependency."
    to: NodeT
    "The inbound node of the edge. This is the dependent."


_NodeMap: TypeAlias = Dict[NodeT, _GraphNodeMeta[NodeT]]
"""A map of node values to the metadata for that node"""


class LowLevelDAG(Generic[NodeT]):
    """
    A generic directed acyclic task-aware-graph. It is a *task-aware* because a
    node can be marked "finished", and one can query the graph for the nodes
    that do not have any un-finished inputs (Meaning the node is "ready").

    :param nodes: The initial nodes of the graph.
    :param edges: The initial edges of the graph.

    .. seealso:: See `.add` for parameter information
    """
    def __init__(self, *, nodes: Iterable[NodeT] = (), edges: Iterable[Edge[NodeT] | tuple[NodeT, NodeT]] = ()) -> None:
        self._nodes: _NodeMap[NodeT] = {}
        'All nodes in the graph'
        self._ready_nodes: set[NodeT] = set()
        'The set of nodes that are ready to be processed'
        self._edges: set[Edge[NodeT]] = set()
        'all of the edges in the graph'
        # Begin by loading up all the initial nodes and edges
        self.add(nodes=nodes, edges=edges)

    def add(self, *, nodes: Iterable[NodeT] = (), edges: Iterable[Edge[NodeT] | tuple[NodeT, NodeT]] = ()) -> None:
        """
        Add content to the graph.

        :param nodes: Nodes to add to the graph.
        :param edges: Edges to add to the graph.

        :raise GraphStateError: If any of the `~Edge.to` nodes in the new
            edges are not `~.NodeState.Pending`. (You cannot add a
            dependency to a running or finished node)
        :raise CycleError: If any of the added edges would create a dependency
            cycle.
        :raise DuplicateEdgeError: If any of the given edges already exist in
            this graph.
        :raise DuplicateNodeError: If any of the given nodes already exist in
            this graph.
        :raise MissingNodeError: If any of the edges refer to nodes that are
            not either already in the graph or added by `.nodes`.

        Edges may be instances of `.Edge` or tuples of node pairs, where
        the first element is the `~.Edge.from_` node and the second element is the
        `~.Edge.to` node.
        """
        # Duplicate the nodes so that we can modify them in-place. Exceptions will leave the
        # graph unchanged.
        ready_nodes = set(self._ready_nodes)
        repl: _NodeMap[NodeT] = {}
        more_edges: set[Edge[NodeT]] = set()

        # First add all the nodes
        for n in nodes:
            if n in self._nodes or n in repl:
                raise DuplicateNodeError(n)
            repl[n] = _GraphNodeMeta(n)
            ready_nodes.add(n)

        # Now add all the edges
        for e in edges:
            if not isinstance(e, Edge):
                e = Edge(e[0], e[1])
            self._add_edge(e, ready_nodes, repl)
            more_edges.add(e)

        # Store the nodes and update the ready-nodes
        self._nodes.update(repl)
        self._edges |= more_edges
        self._ready_nodes = ready_nodes

    def copy(self) -> LowLevelDAG[NodeT]:
        """
        Create a detached clone of this graph. This will copy the node metadata
        and edge metadata, but the nodes themselves will not be copied.
        """
        return copy.deepcopy(self)

    def mark_running(self, node: NodeT) -> None:
        """
        Mark a node as "running"

        .. warning:: Asserts that `node` is a member of the graph and that
            the node has zero pending inputs.
        """
        gnode = self._nodes.get(node)
        assert gnode is not None, ('mark_running() a node which is not in the graph', node)
        assert gnode.pending_inputs == 0, ('mark_running() a node that still has pending inputs', gnode)
        assert gnode.value in self._ready_nodes
        self._ready_nodes.remove(gnode.value)
        gnode.state = NodeState.Running

    def mark_finished(self, node: NodeT) -> None:
        """
        Mark a node as "finished." This will update the ready-nodes of this graph.

        .. note:: It is safe to mark a node as "finished" without having first
            marked it as running with `.mark_running`.

        .. warning:: Asserts that `node` is a member of the graph, that the
            node has no pending inputs, and that the node has not already been
            marked as "finished."
        """
        gnode = self._nodes.get(node)
        assert gnode is not None, ('mark_finished() a node which is not in the graph', node)
        assert gnode.pending_inputs == 0, ('mark_finished() a node that still has pending inputs', gnode)
        assert gnode.state is not NodeState.Finished, ('mark_finished() a node that was already marked as finished',
                                                       gnode)
        if gnode.state is NodeState.Pending:
            #: A node can be marked as finished if it was pending but not running, just skip that state
            self._ready_nodes.remove(gnode.value)
        gnode.state = NodeState.Finished
        for out in gnode.outs:
            out_node = self._nodes.get(out)
            assert out_node is not None, 'Malformed DAG'
            assert node in out_node.ins
            assert out_node.pending_inputs > 0
            out_node.pending_inputs -= 1
            if out_node.pending_inputs == 0:
                # Finishing this node has caused a dependency to become ready
                self._ready_nodes.add(out_node.value)

    @property
    def ready_nodes(self) -> Iterable[NodeT]:
        """The node values in the graph that have no unfinished inputs"""
        return iter(self._ready_nodes)

    @property
    def all_nodes(self) -> Iterable[NodeT]:
        """All nodes currently in the graph"""
        return self._nodes.keys()

    @property
    def all_edges(self) -> Iterable[Edge[NodeT]]:
        """Iterate all the edges in the graph"""
        return iter(self._edges)

    @property
    def has_ready_nodes(self) -> bool:
        """Boolean property on whether the graph has any ready nodes."""
        return bool(self._ready_nodes)

    def dependencies_of(self, node: NodeT) -> Iterable[NodeT]:
        """Iterate the direct dependencies of the given node"""
        return iter(self._nodes[node].ins)

    def dependents_of(self, node: NodeT) -> Iterable[NodeT]:
        """Iterate of the the direct dependents of the given node."""
        return iter(self._nodes[node].outs)

    def state_of(self, node: NodeT) -> NodeState:
        """Get the `.NodeState` of the given `node`"""
        return self._nodes[node].state

    def _add_edge(self, edge: Edge[NodeT], ready_nodes: set[NodeT], repl: _NodeMap[NodeT]) -> None:
        """
        Introduce a new edge to the graph. This requires that all output nodes be pending.

        'repl' is the map of nodes that is being updated and will replace our own map of nodes once
        the update is safely complete.
        """
        from_node = repl.get(edge.from_) or copy.deepcopy(self._nodes.get(edge.from_))
        to_node = repl.get(edge.to) or copy.deepcopy(self._nodes.get(edge.to))

        if edge.from_ == edge.to:
            # Edge points to itself
            raise CycleError([edge.to, edge.to])
        if from_node is None:
            # from-node doesn't exist
            raise MissingNodeError(edge.from_)
        if to_node is None:
            # to-node doesn't exist
            raise MissingNodeError(edge.to)

        if to_node.state is not NodeState.Pending:
            # to-node has already been marked finished/running, so the dependency is broken
            raise GraphStateError(to_node.value)

        if edge.to in from_node.outs:
            # The to-node is already connected to the from-node
            raise DuplicateEdgeError(edge)

        repl[from_node.value] = from_node
        repl[to_node.value] = to_node

        assert edge.from_ not in to_node.ins

        # Connect input and output nodes
        from_node.outs.add(edge.to)
        to_node.ins.add(edge.from_)

        # Recalculate the order of the from-node
        if to_node.order >= from_node.order:
            from_node.order = to_node.order + 1
            # Update the order number. This will also detect and raise on graph cycles
            self._propagate_order(repl, from_node, set(), [from_node.value])

        # If the from-node is not already marked as "finished", then the to-node
        # has an additional pending input
        if from_node.state is not NodeState.Finished:
            to_node.pending_inputs += 1
            if to_node.pending_inputs == 1:
                # We made the node un-ready
                ready_nodes.remove(to_node.value)

    def _get_node(self, repl: _NodeMap[NodeT], k: NodeT) -> _GraphNodeMeta[NodeT]:
        n = repl.get(k)
        if n:
            return n
        n = self._nodes.get(k)
        assert n
        return n

    def _propagate_order(self, repl: _NodeMap[NodeT], node: _GraphNodeMeta[NodeT], visited: set[NodeT],
                         stack: list[NodeT]) -> None:
        pending = (n for n in node.ins if n not in visited)
        for inp in pending:
            visited.add(inp)
            from_node = self._get_node(repl, inp)
            self._recalc_order(repl, from_node, visited, stack)

    def _recalc_order(self, repl: _NodeMap[NodeT], node: _GraphNodeMeta[NodeT], visited: set[NodeT],
                      stack: list[NodeT]) -> None:
        if node.value == stack[0]:
            raise CycleError(stack)
        out_max = max(self._get_node(repl, n).order for n in node.outs)
        new_order = max(out_max + 1, node.order)
        assert new_order >= node.order
        if new_order > node.order:
            n = repl[node.value] = copy.deepcopy(node)
            n.order = new_order
            stack.append(node.value)
            self._propagate_order(repl, node, visited, stack)
            stack.pop()

    def __contains__(self, value: NodeT) -> bool:
        "Check whether the named value is already a node in the graph"
        return value in self._nodes


class DAGView(Generic[NodeT]):
    """
    Obtain a readonly view of a `.LowLevelDAG`.

    :param graph: The graph to view.
    """
    def __init__(self, graph: LowLevelDAG[NodeT]) -> None:
        self._graph = graph

    def dependencies_of(self, node: NodeT) -> Iterable[NodeT]:
        "Get the dependencies of the given node"
        return self._graph.dependencies_of(node)

    def dependents_of(self, node: NodeT) -> Iterable[NodeT]:
        "Get the dependents of the given node"
        return self._graph.dependents_of(node)

    @property
    def all_nodes(self) -> Iterable[NodeT]:
        "Iterate all the nodes in the graph"
        return self._graph.all_nodes

    @property
    def edges(self) -> Iterable[Edge[NodeT]]:
        "Iterate all the edges in the graph"
        return self._graph.all_edges



import random

import pytest

from .. import db as db_mod
from .. import persist as mod
from ..task.task_test import run_test_on_fun


@pytest.fixture
def db() -> db_mod.Database:
    return db_mod.Database.get_or_create(':memory:')


@pytest.mark.asyncio
async def test_persist_global(db: db_mod.Database) -> None:
    with mod.set_global_persistor(mod.DatabasePersistor(db, None)):
        assert not mod.globl.has('eggs')
        mod.globl.set('eggs', 22)
        assert mod.globl.get('eggs') == 22

    with mod.set_global_persistor(mod.DatabasePersistor(db, None)):
        # Reopen the same persistor
        assert mod.globl.has('eggs')
        assert mod.globl.get('eggs') == 22

        # Modify it again
        mod.globl.set('eggs', 42)
        assert mod.globl.get('eggs') == 42

    with mod.set_global_persistor(mod.DatabasePersistor(db, None)):
        assert mod.globl.get('eggs') == 42


def test_persist_local(db: db_mod.Database) -> None:
    tid1 = db.get_task_rowid('example1')
    tid2 = db.get_task_rowid('example2')

    with mod.set_task_local_persistor(mod.DatabasePersistor(db, tid1)):
        assert not mod.local.has('eggs')
        mod.local.set('eggs', 12)
        assert mod.local.get('eggs') == 12

    with mod.set_task_local_persistor(mod.DatabasePersistor(db, tid2)):
        assert not mod.local.has('eggs')
        mod.local.set('eggs', 72)
        assert mod.local.get('eggs') == 72

    with mod.set_task_local_persistor(mod.DatabasePersistor(db, tid1)):
        # Restored from the prior task ID
        assert mod.local.has('eggs')
        assert mod.local.get('eggs') == 12


def test_task_persist_global() -> None:
    val = random.randint(0, 99999)

    async def run() -> None:
        mod.globl.set('test_value', val)

    async def run2() -> None:
        assert mod.globl.get('test_value') == val

    run_test_on_fun(run)
    run_test_on_fun(run2)
    run_test_on_fun(run)
    run_test_on_fun(run2)
    run_test_on_fun(run2)


def test_task_persist_local() -> None:
    val1 = random.randint(0, 5_000)
    val2 = random.randint(5_001, 10_000)

    did_run1 = False

    async def run1() -> None:
        nonlocal did_run1
        if did_run1:
            assert mod.local.get('value') == val1
        else:
            mod.local.set('value', val1)
        did_run1 = True

    did_run2 = False

    async def run2() -> None:
        nonlocal did_run2
        if did_run2:
            assert mod.local.get('value') == val2
            assert mod.local.get('value') != val1
        else:
            mod.local.set('value', val2)
        did_run2 = True

    run_test_on_fun(run1)
    assert did_run1
    run_test_on_fun(run1)

    run_test_on_fun(run2)
    assert did_run2
    run_test_on_fun(run2)



from typing import AsyncContextManager
from typing_extensions import Protocol, runtime_checkable
from dagon.ext.iface import OpaqueTaskGraphView
from dagon.task.dag import OpaqueTask
from dagon.ui.events import UIEvents


@runtime_checkable
class I_UIExtension(Protocol):
    @property
    def dagon_ui_name(self) -> str:
        ...

    @property
    def dagon_ui_opt_name(self) -> str:
        ...

    def ui_global_context(self, graph: OpaqueTaskGraphView, events: UIEvents) -> AsyncContextManager[None]:
        ...

    def ui_task_context(self, task: OpaqueTask) -> AsyncContextManager[None]:
        ...



from __future__ import annotations

import pytest

from . import ll_dag

StringGraph = ll_dag.LowLevelDAG[str]


def test_create_empty():
    g = StringGraph()
    assert set(g.all_nodes) == set()
    assert set(g.ready_nodes) == set()
    assert 'foo' not in g


def test_add_item():
    g = StringGraph(nodes=('foo', 'bar'), edges=[
        ('foo', 'bar'),
    ])
    assert 'foo' in g
    assert 'bar' in g
    assert set(g.ready_nodes) == {'foo'}


def test_finish_item():
    g = StringGraph(nodes=['foo', 'bar'], edges=[
        ('foo', 'bar'),
    ])
    g.mark_finished('foo')
    assert set(g.ready_nodes) == {'bar'}


def test_finish_item_two_chain():
    g = StringGraph(nodes={'foo', 'bar', 'baz'}, edges=[
        ('foo', 'bar'),
        ('bar', 'baz'),
    ])
    g.mark_finished('foo')
    assert set(g.ready_nodes) == {'bar'}
    g.mark_finished('bar')
    assert set(g.ready_nodes) == {'baz'}
    g.mark_finished('baz')
    assert set(g.ready_nodes) == set()


def test_finish_item_multi_inputs():
    g = StringGraph(nodes=('foo', 'bar', 'baz'), edges=[
        ('foo', 'baz'),
        ('bar', 'baz'),
    ])
    assert set(g.ready_nodes) == {'foo', 'bar'}
    g.mark_finished('foo')
    assert set(g.ready_nodes) == {'bar'}
    g.mark_finished('bar')
    assert set(g.ready_nodes) == {'baz'}


def test_finish_item_multi_outputs():
    g = StringGraph(nodes=('foo', 'bar', 'baz'), edges=[
        ('foo', 'bar'),
        ('foo', 'baz'),
    ])
    g.mark_finished('foo')
    assert set(g.ready_nodes) == {'bar', 'baz'}


def test_finish_item_cross():
    g = StringGraph(nodes=('foo', 'bar', 'baz', 'quux'),
                    edges=[
                        ('foo', 'bar'),
                        ('foo', 'baz'),
                        ('quux', 'bar'),
                        ('quux', 'baz'),
                    ])
    assert set(g.ready_nodes) == {'foo', 'quux'}
    g.mark_finished('foo')
    assert set(g.ready_nodes) == {'quux'}
    g.mark_finished('quux')
    assert set(g.ready_nodes) == {'bar', 'baz'}


def test_detect_cycles():
    g = StringGraph(nodes=('foo', 'bar', 'baz', 'quux'), edges=[
        ('foo', 'bar'),
        ('foo', 'baz'),
        ('baz', 'quux'),
    ])
    with pytest.raises(ll_dag.CycleError, match="['quux', 'baz', 'quux']"):
        g.add(edges=[('quux', 'baz')])
    with pytest.raises(ll_dag.CycleError, match="['quux', 'baz', 'foo', 'quux']"):
        g.add(edges=[('quux', 'foo')])
    g.add(edges=[('quux', 'bar')])


def test_detect_cycles_2():
    g = StringGraph(nodes=['foo', 'bar', 'baz'])
    g.add(edges=[
        ('foo', 'bar'),
        ('bar', 'baz'),
    ])
    with pytest.raises(ll_dag.CycleError):
        g.add(edges=[('baz', 'foo')])
    with pytest.raises(ll_dag.CycleError):
        g.add(edges=[('bar', 'foo')])


def test_detect_not_cycle():
    g = StringGraph(nodes=['foo', 'bar', 'baz', 'quux'], edges=[
        ('foo', 'bar'),
        ('baz', 'quux'),
    ])

    g.add(edges=[('bar', 'baz')])


def test_self_dep():
    g = StringGraph()
    with pytest.raises(ll_dag.CycleError, match="['foo', 'foo']"):
        g.add(edges=[('foo', 'foo')])


def test_add_edge_on_finished_error():
    g = StringGraph(nodes=['foo', 'bar'], edges=[('foo', 'bar')])
    assert set(g.ready_nodes) == {'foo'}
    g.mark_finished('foo')
    g.add(nodes=['baz'])
    assert set(g.ready_nodes) == {'bar', 'baz'}
    with pytest.raises(ll_dag.GraphStateError, match='foo'):
        g.add(edges=[('baz', 'foo')])


def test_add_dup_edge_error():
    g = StringGraph(nodes=['foo', 'bar'])
    g.add(edges=[('foo', 'bar')])
    assert set(g.ready_nodes) == {'foo'}
    with pytest.raises(ll_dag.DuplicateEdgeError, match="from_='foo', to='bar'"):
        g.add(edges=[('foo', 'bar')])


def test_add_dup_node_error():
    g = StringGraph(nodes=['foo'])
    with pytest.raises(ll_dag.DuplicateNodeError, match='foo'):
        g.add(nodes=['foo'])


def test_missing_node_error():
    g = StringGraph(nodes=['foo', 'bar'])
    with pytest.raises(ll_dag.MissingNodeError, match='quux'):
        g.add(edges=[('quux', 'bar')])


def test_random_large_graph():
    import random
    items: list[str] = ['foo', 'bar', 'baz']
    g = StringGraph(nodes=items)
    count = 1000
    for v in range(count):
        g.add(nodes=[str(v)])
        items.append(str(v))

    for _ in range(count):
        first = random.randrange(0, len(items) - 4)
        second = random.randrange(first + 1, len(items) - 1)
        try:
            g.add(edges=[(items[first], items[second])])
        except ll_dag.DuplicateEdgeError:
            pass


def test_restore():
    g = StringGraph()
    assert set(g.ready_nodes) == set()
    with pytest.raises(ll_dag.DuplicateNodeError):
        g.add(nodes=['foo', 'bar', 'bar', 'baz'])
    assert set(g.ready_nodes) == set()
    assert list(g.all_nodes) == []

    g.add(nodes=['foo', 'bar', 'baz'], edges=[('foo', 'bar'), ('bar', 'baz')])
    assert set(g.ready_nodes) == {'foo'}
    with pytest.raises(ll_dag.CycleError):
        g.add(edges=[('baz', 'foo')])
    assert set(g.ready_nodes) == {'foo'}


def test_copy():
    g = StringGraph(nodes=['foo', 'baz', 'bar'], edges=[('foo', 'bar'), ('bar', 'baz')])
    assert set(g.ready_nodes) == {'foo'}
    g2 = g.copy()
    g2.mark_running('foo')
    assert set(g.ready_nodes) == {'foo'}
    assert set(g2.ready_nodes) == set()
    g2.mark_finished('foo')
    assert set(g.ready_nodes) == {'foo'}
    assert set(g2.ready_nodes) == {'bar'}
    g2.mark_finished('bar')
    assert set(g.ready_nodes) == {'foo'}
    assert set(g2.ready_nodes) == {'baz'}
    g.add(nodes=['quux'], edges=[('baz', 'quux')])
    assert set(g.all_nodes) == {'foo', 'bar', 'baz', 'quux'}
    assert set(g2.all_nodes) == {'foo', 'bar', 'baz'}


def test_sameness():
    import dagon.core.ll_dag
    assert dagon.core.ll_dag.LowLevelDAG is ll_dag.LowLevelDAG  # type: ignore



"""
Module ``dagon.script_mode``
############################

.. note::
    This is a special module that should only be imported by Python scripts
    that are executed directly.

This module can be used to generate standalone scripts that present as
Dagon-based DAG executors.

Usage is simple::

    from dagon import script_mode

    # ... Task definitions ...

    script_mode.run()

The :func:`run` function will not return, it will raise :class:`SystemExit`
with the result.

.. note::
    The ``dagon.script_mode`` module must be imported **before** any task
    definitions. Importing this module has side-effects, and it should not be
    used outside of Python executable scripts.
"""

from __future__ import annotations

import sys
from contextlib import ExitStack
from typing import NoReturn, Sequence

from dagon.task import dag
from dagon.tool import main
from ..util.doc import __sphinx_build__

if not __sphinx_build__:
    try:
        dag.current_dag()
    except RuntimeError:
        dag.set_current_dag(dag.TaskDAG('<script-mode>'))

    _EXTENSIONS = main.get_extensions()
    _ST = ExitStack()
    _ST.enter_context(_EXTENSIONS.app_context())


def run(argv: Sequence[str] | None = None, *, default_tasks: Sequence[str] | None = None) -> NoReturn:
    """
    Run Dagon using the tasks defined in the script. Raises
    :class:`SystemExit` based on the result of task execution. This call
    should be the final substantial line in the script file, as no following
    code will be executed.

    :param argv: Set the command line arguments to Dagon. If not provided, will
        use :obj:`sys.argv`.
    :param default_tasks: Set the default tasks to execute.

    :raises SystemExit: Unconditionally.
    """
    with _ST:
        sys.exit(main.run_for_dag(
            dag.current_dag(),
            _EXTENSIONS,
            argv=argv,
            default_tasks=default_tasks,
        ))



from __future__ import annotations
import os
import sys

import builtins
import argparse
import warnings
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager
from typing import Any, AsyncIterator, Awaitable

from dagon.core.result import NodeResult
from dagon.ext import loader
from dagon.ui.events import ProgressInfo, UIEvents
from dagon.ui.message import Message, MessageType

from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..task.dag import OpaqueTask
from ..util import AsyncNullContext, Opaque, ReadyAwaitable, unused
from .iface import I_UIExtension
from .proc import PrintProcessResultUIInfo

if sys.version_info < (3, 8):
    from importlib_metadata import EntryPoint
else:
    from importlib.metadata import EntryPoint


class UILoadWarning(Warning):
    "Warning type for when warnings occur while loading a UI extension"


_EVENTS = UIEvents()


class _NullUI:
    dagon_ui_name = 'null'
    dagon_ui_opt_name = 'none'

    def _on_print(self, message: Message) -> None:
        builtins.print(message.content + '\n')

    @asynccontextmanager
    async def ui_global_context(self, graph: OpaqueTaskGraphView, events: UIEvents):
        with ExitStack() as st:
            st.enter_context(events.message.connect(self._on_print))
            yield

    def ui_task_context(self, task: OpaqueTask):
        return AsyncNullContext()


class _Ext(BaseExtension[None, I_UIExtension, None]):
    dagon_ext_name = 'dagon.ui'
    dagon_ext_requires = ['dagon.events']

    def __init__(self) -> None:
        eps = loader.get_entry_points('dagon.uis')
        self._uis: dict[str, I_UIExtension] = {}
        self._chosen_ui: str | None = None
        for ep in eps:
            ui_inst = self._try_load_one(ep)
            if ui_inst:
                self._uis[ui_inst.dagon_ui_opt_name] = ui_inst

    @staticmethod
    def _try_load_one(ep: EntryPoint) -> None | I_UIExtension:
        ep_name: str = getattr(ep, 'name', '<unnamed>')
        try:
            cls: Opaque = ep.load()
        except BaseException as e:
            warnings.warn(f"An exception occurred while trying to load Dagon UI extension {ep!r}: {e}",
                          UILoadWarning,
                          source=e)
            return
        if not hasattr(cls, 'dagon_ui_name'):
            warnings.warn(f'UI extension object {cls!r} does not have a "dagon_ui_name" attribute', UILoadWarning)
            return
        ui_name = getattr(cls, 'dagon_ui_name')
        if not isinstance(ui_name, str):
            warnings.warn(f'UI extension object {cls!r} has a non-string "dagon_ui_name": {ui_name!r}', UILoadWarning)
            return
        if ui_name != ep_name:
            warnings.warn(
                f'UI Extension object {cls!r} has a mismatching UI name. Expected "{ep_name}", but got "{ui_name}"',
                UILoadWarning)
            return
        if not callable(cls):
            warnings.warn(f'UI extension entrypoint {cls!r} is not callable', UILoadWarning)
            return
        try:
            inst: Any = cls()
        except BaseException as e:
            warnings.warn(f'Instantiating/calling UI extension loader [{cls!r}] resulted in an exception',
                          UILoadWarning,
                          source=e)
            return
        if not isinstance(inst, I_UIExtension):
            warnings.warn(f'Generation UI extension object {inst!r} does not implement the necessary interface')
            return
        return inst

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        grp = arg_parser.add_argument_group('UI Options')
        grp.add_argument('--interface',
                         '-ui',
                         help='Set the user inferface kind',
                         choices=['auto', 'none'] + [u.dagon_ui_opt_name for u in self._uis.values()],
                         default='auto')
        return super().add_options(arg_parser)

    def handle_options(self, opts: argparse.Namespace) -> None:
        self._chosen_ui = opts.interface

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[I_UIExtension]:
        chosen = self._chosen_ui
        iface: I_UIExtension
        if chosen is None or chosen == 'none':
            iface = _NullUI()
        else:
            if chosen == 'auto':
                chosen = 'fancy' if os.isatty(sys.__stdout__.fileno()) else 'simple'
            iface = self._uis[chosen]
        async with AsyncExitStack() as st:
            await st.enter_async_context(iface.ui_global_context(graph, _EVENTS))
            yield iface

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask):
        iface = self.global_data()
        async with AsyncExitStack() as st:
            await st.enter_async_context(iface.ui_task_context(task))
            yield

    def notify_result(self, result: NodeResult[OpaqueTask]) -> Awaitable[None]:
        _EVENTS.task_result.emit(result)
        return ReadyAwaitable(None)


def status(message: str) -> None:
    _EVENTS.status.emit(message)


def print(message: str, *, type: MessageType = MessageType.Print) -> None:
    _EVENTS.message.emit(Message(message, type))


def progress(value: float | None) -> None:
    _EVENTS.progress.emit(ProgressInfo(value))


def print_process_done(result: PrintProcessResultUIInfo) -> None:
    _EVENTS.print_process_done.emit(result)


unused(_Ext)



"""
Execution options
"""

from __future__ import annotations

import enum
from pathlib import Path
from typing import (Any, Callable, Generic, Mapping, Type, TypeVar, Union, cast, overload)

from typing_extensions import Protocol

from ..util import T, U, Undefined, UndefinedType


class _CustomDagonOpt(Protocol):
    @classmethod
    def __dagon_parse_opt__(cls, val: str) -> _CustomDagonOpt:
        ...


SimpleOptionType = Union[str, bool, Path, int, float, enum.Enum, _CustomDagonOpt]
SimpleOptionT = TypeVar('SimpleOptionT', bound=SimpleOptionType)


def convert_bool(s: str) -> bool:
    """
    Convert the given string into a bool with some common bool-ish values.
    """
    b = {
        '0': False,
        'no': False,
        'n': False,
        'false': False,
        '1': True,
        'yes': True,
        'y': True,
        'true': True,
    }.get(s.lower())
    if b is None:
        raise ValueError(f'Invalid boolean option string "{s}"')
    return b


_CONVERSION_MAP: Mapping[Type[Any], Callable[[str], Any]] = {
    str: str,
    bool: convert_bool,
    Path: Path,
    int: int,
    float: float,
}

_PARSE_OPT_METHOD_KEY = '__dagon_parse_opt__'


class NoSuchOptionError(NameError):
    pass


def is_valid_option_type(typ: Type[Any]) -> bool:
    """
    Return `True` if the given type is valid for use as a type with :class:`Option`.
    """
    return hasattr(typ, _PARSE_OPT_METHOD_KEY) or issubclass(typ, enum.Enum) or (typ in _CONVERSION_MAP)


def parse_value_str(typ: Type[SimpleOptionT], val_str: str) -> SimpleOptionT:
    """
    Given a type `typ` and a value string `val_str`, attempt to convert
    that string to the given `typ`. Raises `ValueError` in cases of
    failure.
    """
    try:
        conv = getattr(type, _PARSE_OPT_METHOD_KEY, None)
        if conv is not None:
            assert callable(conv)
            return cast(SimpleOptionT, conv(val_str))  # pylint: disable=not-callable
        if issubclass(typ, enum.Enum):
            try:
                return cast(SimpleOptionT, typ(val_str))
            except ValueError as e:
                try:
                    i = int(val_str)
                except ValueError:
                    raise e from e
                else:
                    return cast(SimpleOptionT, typ(i))
        return cast(SimpleOptionT, _CONVERSION_MAP[typ](val_str))
    except Exception as e:
        raise ValueError(f'Failed to parse "{val_str}" as type `{typ.__name__}`') from e


def get_type_parser(typ: Type[SimpleOptionT]) -> Callable[[str], SimpleOptionT]:
    return lambda s: parse_value_str(typ, s)


class Option(Generic[T]):
    """
    Representation of a run-time execution option that can be provided by a
    user.

    :param name: The name of the option. Should be unique in a given Dag.
    :param typ: The type of the option. Must be a subtype of `.OptionType`.
    :param default: The default value, or a factory function that generates the
        default value. If omitted and no value was provided by a user,
        requesting the option's value will return `None`.
    :param doc: A documentation string to display to the user.
    :param validate: A secondary validation function that checks that a given
        value is valid for this option. Should return `None` when the value
        is valid, or an error string explaining why it is an invalid value.
    """
    @overload
    def __init__(self,
                 name: str,
                 *,
                 type: Type[T] | UndefinedType,
                 parse: Callable[[str], T],
                 doc: str | None = None,
                 validate: Callable[[T], str | None] | None = None) -> None:
        ...

    @overload
    def __init__(self,
                 name: str,
                 *,
                 type: Type[T] | UndefinedType,
                 parse: Callable[[str], T],
                 default: T = ...,
                 doc: str | None = None,
                 validate: Callable[[T], str | None] | None = None) -> None:
        ...

    @overload
    def __init__(self,
                 name: str,
                 *,
                 type: Type[T] | UndefinedType,
                 parse: Callable[[str], T],
                 calc_default: Callable[[], T] = ...,
                 doc: str | None = None,
                 validate: Callable[[T], str | None] | None = None) -> None:
        ...

    def __init__(self,
                 name: str,
                 *,
                 type: Type[T] | UndefinedType = Undefined,
                 parse: Callable[[str], T],
                 calc_default: Callable[[], T] | None = None,
                 doc: str | None = None,
                 validate: Callable[[T], str | None] | None = None,
                 **kwargs: Any) -> None:
        self._name = name
        self._parser = parse
        self._type = type
        self._calc_default = calc_default
        self._validate = validate
        self._doc = doc
        self._has_default = 'default' in kwargs
        self._default: T | UndefinedType = kwargs.pop('default', Undefined)

    @property
    def name(self) -> str:
        """The option's name"""
        return self._name

    @property
    def doc(self) -> str | None:
        """Documentation for this option"""
        return self._doc

    @overload
    def get(self) -> T:
        ...

    @overload
    def get(self, *, default: U) -> T | U:
        ...

    def get(self, **kw: Any) -> T:
        """
        Obtain the value that was given to this option.

        .. note::
            This method may only be called from within a task-executing context!
        """
        from dagon.option.ext import ctx_fulfilled_options
        return ctx_fulfilled_options().get(self, **kw)

    @property
    def type(self) -> Type[T] | UndefinedType:
        return self._type

    @property
    def has_default(self):
        """Whether the option has a default value provided"""
        return self._has_default

    def get_default(self) -> T:
        """
        Get the default value associated with this option, or :data:`Undefined`
        if no default is provided.
        """
        if self._calc_default is not None:
            return self._calc_default()
        return cast(T, self._default)

    def validate(self, value: T) -> str | None:
        """
        Check if the given value is valid for this option.

        Returns a string representing the error, or `None` if there was
        no error.
        """
        if self._validate is not None:
            return self._validate(value)
        return None

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: Any) -> bool:
        return self is other

    def __repr__(self) -> str:
        return f'<dagon.option.Option "{self.name}">'

    def parse_str(self, val_str: str) -> T:
        """
        Parse the given string into a value for this option. Does not perform
        additional validation.
        """
        return self._parser(val_str)



import argparse
from pathlib import Path
import datetime
import enum
import sys
import textwrap
from typing import NoReturn, Sequence, Optional, Iterable
from string import Template
from typing_extensions import TypeAlias

from dagon.db import Database, FileInfo, RunID
from dagon.ui import ansi
from dagon.util import first, unused
from . import graph, timeline


class QueryChoice(enum.Enum):
    Run = 'run'
    AllRuns = 'all-runs'
    Proc = 'proc'
    Files = 'files'
    Graph = 'graph'
    Timeline = 'timeline'


_RUN_INFO_TEMPLATE = Template(r'''Run ID: $run_id
Start time: $start_time_iso ($start_time_human)
''')

_TASK_RUN_REPORT = Template(r'''Task: $task_name
  Start time: $start_time_iso ($start_time_human)
    Duration: ${duration} sec
      Result: $result
''')

_SUBPROC_EXEC_REPORT = Template(r'''External process: $exec_id
  Command: $process_cmdline
  Start Directory: $cwd
  Start time: $start_time_iso ($start_time_human)
  Duration: $duration sec
  Exit code: $retc
  stdout: $stdout_what
  stderr: $stderr_what
''')


def make_pretty_date(dt: datetime.datetime) -> str:
    # Python 3.7 adds a `fromisoformat`. Use that after upgrading.
    return dt.strftime('%A, %B %d, %I:%M:%S %p')


_RunTup: TypeAlias = 'tuple[RunID, float]'
_ProcTup: TypeAlias = 'tuple[int, int, str, str, str, str, int, int, float]'


def _print_run_info(row_tup: _RunTup) -> None:
    run_id, start_time_ = row_tup
    start_time = datetime.datetime.fromtimestamp(start_time_)

    report = _RUN_INFO_TEMPLATE.substitute(run_id=run_id,
                                           start_time_iso=start_time.isoformat(),
                                           start_time_human=make_pretty_date(start_time))
    print(report)


def _print_proc(proc_tup: _ProcTup, indent: str = '') -> None:
    exec_id, _task_run_id, cmd, cwd, stdout, stderr, retc, start_time_, duration = proc_tup
    unused(_task_run_id)
    stdout_what = f'{len(stdout)} bytes' if stdout is not None else 'Not attached'
    stderr_what = f'{len(stderr)} bytes' if stderr is not None else 'Not attached'
    start_time = datetime.datetime.fromtimestamp(start_time_)
    report = _SUBPROC_EXEC_REPORT.substitute(exec_id=exec_id,
                                             process_cmdline=cmd,
                                             cwd=cwd,
                                             start_time_iso=start_time.isoformat(),
                                             start_time_human=make_pretty_date(start_time),
                                             retc=retc,
                                             duration=duration,
                                             stdout_what=stdout_what,
                                             stderr_what=stderr_what)
    report = textwrap.indent(textwrap.dedent(report), indent)
    print(report)


def _print_proc_execs(_args: argparse.Namespace, db: Database, task_run_id: int, indent: str = '') -> None:
    execs = db('SELECT * FROM dagon_proc_execs WHERE task_run_id = :tri ORDER BY start_time', tri=task_run_id)
    for proc in execs:
        _print_proc(proc, indent)


def _run_query(runid: Optional[int], db: Database) -> Optional[_RunTup]:
    rows: Iterable[_RunTup] = db(
        r'''
        SELECT run_id, time
        FROM dagon_runs
        WHERE run_id=:run OR :run IS NULL
        ORDER BY run_id DESC
        ''',
        run=runid,
    )
    return first(rows, default=None)


def _print_run_query(args: argparse.Namespace, db: Database) -> int:
    run_tup = _run_query(args.run, db)
    if not run_tup:
        raise RuntimeError('Database has no runs')

    _print_run_info(run_tup)
    run_id = run_tup[0]

    rows = db('''SELECT task_run_id,
                   task_id,
                   run_id,
                   end_state,
                   start_time,
                   duration,
                   name
                FROM dagon_task_runs
                JOIN dagon_tasks USING(task_id)
                WHERE run_id=:rid
                ORDER BY duration DESC
            ''',
              rid=run_id)
    for task_run_id, _target_id, _run_id, end_state, start_time_, duration_s, task_name in rows:
        unused(_target_id, _run_id)
        color_fn = {
            'succeeded': ansi.bold_green,
            'failed': ansi.bold_red,
            'cancelled': ansi.bold_yellow,
            'pending': ansi.bold_cyan,
            'running': ansi.bold_cyan,
        }.get(end_state, lambda s: s)
        colored_res = color_fn(end_state.upper())
        start_time = datetime.datetime.fromtimestamp(start_time_)
        report = _TASK_RUN_REPORT.substitute(
            task_name=task_name,
            start_time_iso=start_time.isoformat(),
            start_time_human=make_pretty_date(start_time),
            duration=duration_s,
            result=colored_res,
        )
        print(report)
        _print_proc_execs(args, db, task_run_id, indent='  ')

    return 0


def _print_all_runs(_args: argparse.Namespace, db: Database) -> int:
    for row in db('SELECT * FROM dagon_runs'):
        _print_run_info(row)
    return 0


def _print_proc_output(args: argparse.Namespace, db: Database) -> int:
    if not args.proc_id:
        raise RuntimeError('Pass a --proc-id to query')
    rows = list(db('SELECT * FROM dagon_proc_execs WHERE proc_exec_id=:id', id=args.proc_id))
    if not len(rows):
        raise RuntimeError(f'No subprocess with the given ID {args.proc_id}')
    tup = rows[0]
    assert len(rows) == 1
    _print_proc(tup)
    stdout, stderr = tup[4], tup[5]
    if stdout is not None:
        print('--------------- stdout ---------------')
        print(stdout.decode(), end='')
    if stderr is not None:
        print('--------------- stderr ---------------')
        print(stderr.decode(), end='')
    return 0


def _extract_file(db: Database, file: FileInfo, dest: Path) -> None:
    datas = db.iter_file_data(file)
    dest.parent.mkdir(parents=True, exist_ok=True)
    with dest.open('wb') as fd:
        for blob in datas:
            fd.write(blob)


def _do_files(args: argparse.Namespace, db: Database) -> int:
    dest_dir = None
    ex_pat = args.extract

    if ex_pat:
        dest_dir = Path(args.dest or '.')
        dest_dir.mkdir(parents=True, exist_ok=True)

    run_tup = _run_query(args.run, db)
    if not run_tup:
        raise RuntimeError('Database contains no runs')
    run_id = run_tup[0]
    files = db.iter_files(run_id=run_id)

    if ex_pat is None:
        print(f'Run {run_id} contains the following files:')

    for info in files:
        if ex_pat is None:
            # We're just listing files, not extracting
            size = info.pretty_size if args.human else f'{info.size} bytes'
            print(f'  {info.path} ({size})')
            continue

        if not info.path.match(ex_pat):
            continue

        components = str(info.path).split('/')
        if args.strip_components >= len(components):
            print(f'WARN: Cannot extract {info.path} with --strip-components N >= {len(components)}')
            continue

        dest_relpath = '/'.join(components[args.strip_components:])
        assert dest_dir
        dest = dest_dir / dest_relpath
        print(f'Extract {info.path} to {dest} ... ', end='')
        _extract_file(db, info, dest)
        print('Done')

    return 0


def _do_graph(args: argparse.Namespace, db: Database) -> int:
    if not graph.HAVE_GRAPHVIZ:
        print('Install the Python `graphviz` module to use `graph` inspection', file=sys.stderr)
        return 2
    run_tup = _run_query(args.run, db)
    if not run_tup:
        print('Database contains no runs')
        return 1
    graph.generate_graph(db, run_tup[0])
    return 0


def _do_timeline(args: argparse.Namespace, db: Database) -> int:
    run_tup = _run_query(args.run, db)
    if not run_tup:
        print('Database has no Dagon runs')
        return 1
    html_content = timeline.generate(db, run_tup[0])
    Path(args.out or 'dagon-timeline.html').write_bytes(html_content)
    return 0


def _argparse_run(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.Run)


def _argparse_all_runs(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.AllRuns)


def _argparse_proc(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.Proc)


def _argparse_files(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.Files)
    p.add_argument(
        '--extract',
        '-e',
        nargs='?',
        const='*',
        help='Extract files (with optional glob pattern)',
    )
    p.add_argument('--dest', '-d', help='Destination directory for the files (used with --extract)')
    p.add_argument('--strip-components',
                   '-p',
                   help='Strip leading directory components from source (used with --extract)',
                   type=int,
                   default=0)


def _argparse_graph(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.Graph)


def _argparse_timeline(p: argparse.ArgumentParser) -> None:
    p.set_defaults(choice=QueryChoice.Timeline)
    p.add_argument('--out', '-o', help='Destination path for the HTML timeline')


def main(argv: Sequence[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--database', help='The database to read [.dagon.db]')
    parser.add_argument('--run', help='The run to inspect (default is the latest)', type=int)
    parser.add_argument('--proc-id', help='A subprocess ID (for `proc` mode)', type=int)
    parser.add_argument('--human-readable',
                        '-H',
                        dest='human',
                        help='Prefer human-readable value representatoins',
                        action='store_true')

    sub = parser.add_subparsers()

    _argparse_run(sub.add_parser(QueryChoice.Run.value, help='Request runs'))
    _argparse_all_runs(sub.add_parser(QueryChoice.AllRuns.value, help='Information about all available runs'))
    _argparse_proc(sub.add_parser(QueryChoice.Proc.value, help='Process information'))
    _argparse_files(sub.add_parser(QueryChoice.Files.value, help='Embedded files'))
    _argparse_graph(sub.add_parser(QueryChoice.Graph.value, help='Generate a target dependency graph from a run'))
    _argparse_timeline(sub.add_parser(QueryChoice.Timeline.value, help='Generate an interactive execution timeline'))

    # parser.add_argument(
    #     'query', help='The attribute to query', choices=[v.value for v in QueryChoice])
    args = parser.parse_args(argv)

    db_path = Path(args.database or '.dagon.db').absolute()
    db = Database.get_existing(db_path)

    ansi.ensure_ansi_term()
    if not hasattr(args, 'choice'):
        parser.print_usage()
        return 1

    query = QueryChoice(args.choice)
    fn = {
        QueryChoice.Run: _print_run_query,
        QueryChoice.AllRuns: _print_all_runs,
        QueryChoice.Proc: _print_proc_output,
        QueryChoice.Files: _do_files,
        QueryChoice.Graph: _do_graph,
        QueryChoice.Timeline: _do_timeline,
    }[query]
    return fn(args, db)


def start() -> NoReturn:
    sys.exit(main(sys.argv[1:]))



from __future__ import annotations

from pathlib import Path
import pytest

from dagon import fs

async_test = pytest.mark.asyncio


@async_test
async def test_copy_file(tmp_path: Path) -> None:
    tmp_path.joinpath('foo.txt').write_bytes(b'hello')
    await fs.copy_file(tmp_path / 'foo.txt', tmp_path / 'bar.txt')
    assert tmp_path.joinpath('bar.txt').read_bytes() == b'hello'


@async_test
async def test_copy_noexist_file(tmp_path: Path) -> None:
    with pytest.raises(FileNotFoundError):
        await fs.copy_file(tmp_path.joinpath('nope.txt'), 'nope.txt')


@async_test
async def test_copy_to_noexist_dir(tmp_path: Path) -> None:
    tmp_path.joinpath('foo.txt').write_bytes(b'hi')
    tmp_path.joinpath('bar.txt').write_bytes(b'yo')
    with pytest.raises(FileNotFoundError):
        await fs.copy_file(tmp_path.joinpath('foo.txt'), tmp_path.joinpath('a/b'), mkdirs=False)
    with pytest.raises(NotADirectoryError):
        await fs.copy_file(tmp_path.joinpath('foo.txt'), tmp_path.joinpath('bar.txt/b'), mkdirs=False)
    with pytest.raises(NotADirectoryError):
        await fs.copy_file(tmp_path.joinpath('foo.txt'), tmp_path.joinpath('bar.txt/b'), mkdirs=True)
    # Okay: Creates the directory:
    await fs.copy_file(tmp_path.joinpath('foo.txt'), tmp_path.joinpath('a/b'), mkdirs=True)
    with pytest.raises(IsADirectoryError):
        await fs.copy_file(tmp_path.joinpath('a'), tmp_path.joinpath('b'))


@async_test
async def test_copy_file_is_dir(tmp_path: Path) -> None:
    tmp_path.joinpath('foo.dir').mkdir()
    with pytest.raises(IsADirectoryError):
        await fs.copy_file(tmp_path.joinpath('foo.dir'), tmp_path.joinpath('bar.dir'))


def test_iter_pathish() -> None:
    assert list(fs.iter_pathish('foo')) == [Path('foo')]
    assert list(fs.iter_pathish('foo', type=str)) == ['foo']
    assert list(fs.iter_pathish(['foo'], type=str)) == ['foo']
    assert list(fs.iter_pathish(['foo', 'bar'], type=str)) == ['foo', 'bar']


@async_test
async def test_create_tree(tmp_path: Path) -> None:
    await fs.create_tree(tmp_path, {
        'a.txt': b'Hello!',
        'b.txt': b'Goodbye!',
        'c': {
            'inner.txt': b'Howdy',
        },
    })
    assert tmp_path.joinpath('a.txt').read_bytes() == b'Hello!'
    assert tmp_path.joinpath('b.txt').read_bytes() == b'Goodbye!'
    assert tmp_path.joinpath('c/inner.txt').read_bytes() == b'Howdy'


@async_test
async def test_copy_tree(tmp_path: Path) -> None:
    await fs.create_tree(tmp_path.joinpath('base'), {
        'a': b'a',
        'b': b'b',
        'c': {
            'd': {
                'e': b'e',
            },
        },
    })
    await fs.copy_tree(tmp_path / 'base', tmp_path / 'target')
    assert tmp_path.joinpath('base').is_dir()  # Directory still exists
    assert tmp_path.joinpath('target').is_dir()  # Directory was created
    assert tmp_path.joinpath('target/c/d/e').read_bytes() == b'e'


@async_test
async def test_dir_merge(tmp_path: Path) -> None:
    await fs.create_tree(tmp_path, {
        'a': {
            'file.a.txt': b'foo'
        },
        'b': {
            'file.b.txt': b'bar'
        },
    })
    await fs.copy_tree(tmp_path / 'a', tmp_path / 'c')
    await fs.copy_tree(tmp_path / 'b', tmp_path / 'c', if_exists='merge')
    assert tmp_path.joinpath('c/file.a.txt').read_bytes() == b'foo'
    assert tmp_path.joinpath('c/file.b.txt').read_bytes() == b'bar'
    tmp_path.joinpath('a/file.a.txt').write_bytes(b'baz')
    await fs.copy_tree(tmp_path / 'a', tmp_path / 'c', if_file_exists='keep', if_exists='merge')
    assert tmp_path.joinpath('c/file.a.txt').read_bytes() == b'foo'
    await fs.copy_tree(tmp_path / 'a', tmp_path / 'c', if_file_exists='replace', if_exists='merge')
    assert tmp_path.joinpath('c/file.a.txt').read_bytes() == b'baz'



"""
Module ``dagon.fs``
###################

Module for performing filesystem operations
"""

from __future__ import annotations

import asyncio
import errno
import functools
import itertools
import os
import shutil
from concurrent.futures import Executor, ThreadPoolExecutor
from io import BufferedIOBase
from pathlib import Path, PurePath
from typing import (TYPE_CHECKING, Any, AsyncContextManager, AsyncIterable, AsyncIterator, Awaitable, Callable, Generic,
                    Iterable, Mapping, Type, TypeVar, Union, cast)

from typing_extensions import Literal, ParamSpec

from ..event import CancellationToken, raise_if_cancelled
from ..util import T
from ..util.doc import __sphinx_build__

Pathish = Union['os.PathLike[str]', str]
'A path-able argument. A string, subclass of `~pathlib.PurePath`, or anything with an ``__fspath__`` member'
NPaths = Union[Pathish, Iterable[Pathish]]
'Convenience type to represent either a single path or an iterable list of paths'
FilePredicate = Callable[[Path], bool]
'Type for predicate functions that accept a file path and return a `bool`'

_FS_OPS_POOL = ThreadPoolExecutor(4)

_PathT = TypeVar('_PathT', bound=Pathish)

_Params = ParamSpec('_Params')

if TYPE_CHECKING:

    class _ExecutorMappedOperation(Generic[_Params, T]):
        def __init__(self, sync_func: Callable[_Params, T], executor: Executor) -> None:
            ...

        def sync(self, *args: _Params.args, **kwargs: _Params.kwargs) -> T:
            ...

        def __call__(self, *args: _Params.args, **kwargs: _Params.kwargs) -> Awaitable[T]:
            ...

else:
    _FuncT = TypeVar('_FuncT', bound=Callable[..., Any])

    class _ExecutorMappedOperation(Generic[_FuncT, T]):
        def __init__(self, sync_func: _FuncT, executor: Executor) -> None:
            self.__name__ = sync_func.__name__
            self.__doc__ = sync_func.__doc__
            self._sync = sync_func
            self._exec = executor

        def sync(self, *args: Any, **kwargs: Any) -> T:
            return self._sync(*args, **kwargs)

        def __call__(self, *args: Any, **kwargs: Any) -> Awaitable[T]:
            loop = asyncio.get_running_loop()
            cancel = CancellationToken.get_context_local()
            return loop.run_in_executor(self._exec, lambda: self.sync(*args, **kwargs, cancel=cancel))


def _thread_pooled(fn: Callable[_Params, T]) -> _ExecutorMappedOperation[_Params, T]:
    if __sphinx_build__:
        return cast(Any, fn)
    r = _ExecutorMappedOperation[_Params, T](fn, _FS_OPS_POOL)
    functools.update_wrapper(r, fn)
    return r


def iter_pathish(p: NPaths, *, type: Type[_PathT] = Path) -> Iterable[_PathT]:
    """
    Given an iterable of paths or a single path, yield either the single path
    or each of the paths given.
    """
    if isinstance(p, (str, PurePath)) or hasattr(p, '__fspath__'):
        yield type(p)  # type: ignore
    else:
        yield from (type(i) for i in p)  # type: ignore


IfExists = Literal['replace', 'fail', 'keep']
"""Policy for performing file operations where the destination already exists"""

IfDirectoryExists = Literal['replace', 'fail', 'keep', 'merge']
"""Policy for performing file operations where the destination directory already exists"""


def fs_thread_pool() -> ThreadPoolExecutor:
    """
    The thread pool that is used to execute filesystem operations
    """
    return _FS_OPS_POOL


def _run_fs_op(fn: Callable[[], None]) -> 'asyncio.Future[None]':
    loop = asyncio.get_event_loop()
    return loop.run_in_executor(fs_thread_pool(), fn)  # type: ignore


def _copy_file_no_overwrite(src: Path, dst: Path) -> None:
    with src.open('rb') as f:
        with dst.open('xb') as out:
            shutil.copyfileobj(f, out)


def _any_parent_is_file(p: Path) -> bool:
    "Determine whether 'p' (or any ancestor of 'p') names a file"
    if p.is_dir():
        return False
    if p.is_file():
        return True
    return _any_parent_is_file(p.parent)


@_thread_pooled
def copy_file(file: Pathish,
              dest: Pathish,
              *,
              if_exists: IfExists = 'fail',
              mkdirs: bool = True,
              preserve_symlinks: bool = True,
              preserve_stat: bool = False,
              cancel: CancellationToken | None = None) -> None:
    """
    Copy a file from one location to another.

    :param file: Path to a file to copy.
    :param dest: The full filepath for the destination.
    :param on_exists: What to do if the destination already exists.
    :param mkdirs: If ``True``, make intermediate directories for the destination.
    :param preserve_symlinks: If ``True``, and ``file`` refers to a symlink, the
        symlink will be copied (rather than the target thereof)
    :param preserve_stat: If ``True``, attributes of the source file will be
        applied to the destination.
    :param cancel: A cancellation token to cancel the operation.

    .. note::
        ``dest`` is the full path to a file, not the path to the directory that
        should receive the file.
    """
    raise_if_cancelled(cancel)
    file = Path(file)
    dest = Path(dest)
    pardir = dest.parent
    if not pardir.exists() and mkdirs:
        pardir.mkdir(exist_ok=True, parents=True)

    do_copy: Callable[[Path, Path], Any]
    if preserve_symlinks and file.is_symlink():
        # Copy the link, not the target of the link
        do_copy = lambda s, d: os.symlink(os.readlink(s), d)
    else:
        # Copy file, but raise if we attempt to overwrite it
        do_copy = _copy_file_no_overwrite
    try:
        do_copy(file, dest)
    except FileNotFoundError as e:
        # Windows might not raise NotADirectoryError itself, so we'll check that
        # case ourselves
        if _any_parent_is_file(dest):
            raise NotADirectoryError from e
        raise
    except PermissionError as e:
        if file.is_dir():
            raise IsADirectoryError from e
        raise
    except FileExistsError:
        if if_exists == 'fail':
            raise
        if if_exists == 'keep':
            return
        assert if_exists == 'replace'
        # Delete the file, and do it again.
        dest.unlink()
        do_copy(file, dest)

    if preserve_stat:
        shutil.copystat(file, dest, follow_symlinks=False)
    else:
        shutil.copymode(file, dest, follow_symlinks=False)


@_thread_pooled
def copy_tree(inpath: Pathish,
              destpath: Pathish,
              *,
              mkdirs: bool = True,
              if_exists: IfDirectoryExists = 'fail',
              if_file_exists: IfExists = 'fail',
              only_if: FilePredicate | None = None,
              unless: FilePredicate | None = None,
              preserve_symlinks: bool = True,
              preserve_stat: bool = False,
              cancel: CancellationToken | None = None) -> None:
    """
    Copy a tree file files/directories.

    :param inpath: Path to a file/directory to copy.
    :param destpath: The destination path to copy to.
    :param mkdirs: If ``True``, create intermediate directories.
    :param on_exists: The action to take if the destination directory exists.
    :param on_file_exists: The action to take if any files within the destination
        directory already exists. (This is applicable to ``"merge"``)
    :param only_if: A predicate function that takes an absolute ``Path`` and
        returns a ``bool``. If provided, only files for which this predicate
        returns ``True`` will be copied into the destination.
    :param unless: A predictae function that takes an absolute ``Path`` and
        returns a ``bool``. If provided, only files for which this predicate
        returns ``False`` will be copied into the destination.
    :param preserve_symlinks: Whether to copy symlinks as symlinks, or follow
        them through to copy their targets.
    :param preserve_stat: If ``True``, attributes from the source will be
        applied to the destination.
    :param cancel: A cancellation token for the operation.

    .. note::
        If both ``only_if`` and ``unless`` are provided, ``only_if`` will be
        checked *first*, and then ``unless`` will be checked. Both predicates
        will be considered.
    """
    raise_if_cancelled(cancel)
    indir = Path(inpath).absolute()
    destdir = Path(destpath)
    pardir = destdir.parent
    if not pardir.is_dir():
        if mkdirs:
            pardir.mkdir(exist_ok=True, parents=True)
        else:
            raise RuntimeError(f'Cannot copy directory {indir} to {destdir}: Path {pardir} '
                               'is not an existing directory')
    if destdir.exists():
        if if_exists == 'replace':
            shutil.rmtree(destdir)
        if if_exists == 'fail':
            raise RuntimeError(f'Cannot copy {indir} to {destdir}: Destination already exists')
        if if_exists == 'keep':
            return
        if if_exists == "merge":
            pass

    destdir.mkdir(exist_ok=True, parents=True)
    raise_if_cancelled(cancel)
    for in_abs in indir.iterdir():
        raise_if_cancelled(cancel)
        dest_abs = destdir / (in_abs.relative_to(indir))
        # Item is a symlink, and that symlink refers to a directory, and
        # we are not being asked to preserve symlinks. Resolve the link
        # and copy the tree from the directory:
        should_follow_dir_symlink = (in_abs.is_symlink() and not preserve_symlinks and in_abs.resolve().is_dir())
        # Whether we should copy the tree:
        should_copy_tree = in_abs.is_dir() or should_follow_dir_symlink
        if should_copy_tree:
            copy_tree.sync(in_abs,
                           dest_abs,
                           if_file_exists=if_file_exists,
                           only_if=only_if,
                           unless=unless,
                           preserve_stat=preserve_stat,
                           preserve_symlinks=preserve_symlinks,
                           cancel=cancel)
            # Process the next item
            continue

        # This is a file, not a directory
        if only_if is not None and not only_if(in_abs):
            # The file has been excluded by the `only_if` filter
            continue
        if unless is not None and unless(in_abs):
            # The file has been excluded by the `unless` filter
            continue
        copy_file.sync(in_abs,
                       dest_abs,
                       if_exists=if_file_exists,
                       mkdirs=False,
                       preserve_symlinks=preserve_symlinks,
                       preserve_stat=preserve_stat)


TreeItem = Union[bytes, Mapping[str, 'TreeItem']]


async def create_tree(root: Pathish, items: Mapping[str, TreeItem], *, cancel: CancellationToken | None = None) -> None:
    raise_if_cancelled(cancel)
    return await _run_fs_op(lambda: _create_fs_tree(root, items, cancel))


def _create_fs_tree(root: Pathish, items: Mapping[str, TreeItem], cancel: CancellationToken | None) -> None:
    root = Path(root)
    root.mkdir(exist_ok=True, parents=True)
    for filename, item in items.items():
        raise_if_cancelled(cancel)
        if isinstance(item, bytes):
            root.joinpath(filename).write_bytes(item)
        else:
            _create_fs_tree(root / filename, item, cancel)


def _remove(files: NPaths, recurse: bool, absent_ok: bool) -> None:
    for filepath in iter_pathish(files):
        try:
            if filepath.is_dir() and recurse:
                shutil.rmtree(filepath)
            else:
                os.unlink(filepath)
        except FileNotFoundError:
            if not absent_ok:
                raise


@_thread_pooled
def remove(files: NPaths,
           *,
           recurse: bool = False,
           absent_ok: bool = False,
           cancel: CancellationToken | None = None) -> None:
    """
    Remove a file or directory.

    :param files: Path(s) to an existing files/directories to remove.
    :param recurse: If ``True`` and ``filepath`` names a directory, recursively
        remove the directory as if by :func:`shutil.rmtree`.
    :param absent_ok: If ``True`` and the filepath does not exist, no error will
        be raised. This can create simple idempotent file/directory removal.
    :param cancel: A cancellation token for the operation.
    """
    for filepath in iter_pathish(files):
        raise_if_cancelled(cancel)
        try:
            if filepath.is_dir() and recurse:
                shutil.rmtree(filepath)
            else:
                os.unlink(filepath)
        except FileNotFoundError:
            if not absent_ok:
                raise


@_thread_pooled
def clear_directory(dirs: NPaths, *, cancel: CancellationToken | None = None) -> None:
    """
    Remove the contents of a directory, but keep the directory.

    :param dirs: Path(s) to an directories to clear.
    :param cancel: A cancellation token for the operation.
    """
    raise_if_cancelled(cancel)
    each_dir = iter_pathish(dirs)
    children_per_dir = (d.iterdir() for d in each_dir)
    all_children = itertools.chain.from_iterable(children_per_dir)
    _remove(all_children, recurse=True, absent_ok=True)


def _read_some(fd: BufferedIOBase, buf: bytearray) -> int:
    return fd.readinto(buf)


async def read_blocks_from(io: BufferedIOBase,
                           *,
                           block_size: int = 1024 * 10,
                           cancel: CancellationToken | None = None) -> AsyncIterator[memoryview]:
    buf = bytearray(block_size)
    loop = asyncio.get_event_loop()
    while True:
        raise_if_cancelled(cancel)
        nread = await loop.run_in_executor(_FS_OPS_POOL, lambda: _read_some(io, buf))
        if nread == 0:
            break
        yield memoryview(buf)[:nread]


async def read_blocks(filepath: Pathish,
                      *,
                      block_size: int = 1024 * 10,
                      cancel: CancellationToken | None = None) -> AsyncIterable[memoryview]:
    """
    Asynchronously lazily ready blocks from the given file.
    """
    raise_if_cancelled(cancel)
    filepath = Path(filepath)
    with filepath.open('rb') as fd:
        async for b in read_blocks_from(fd, block_size=block_size, cancel=cancel):
            yield b


@_thread_pooled
def safe_move_file(source: Pathish,
                   dest: Pathish,
                   *,
                   mkdirs: bool = True,
                   if_exists: IfExists = 'fail',
                   cancel: CancellationToken | None = None) -> Path:
    """
    Move the given file from one location to another. If the system cannot
    simply move the file, it will be copied and the source will be removed.

    :param source: The source file which will be moved.
    :param dest: The destination of the move operation.
    :param mkdirs: If ``True``, the directory containing ``dest`` will be created if absent.
    :param on_exists: What to do if the destination already exists.
    :param cancel: A cancellation token for the operation.
    """
    source = Path(source)
    dest = Path(dest)
    raise_if_cancelled(cancel)
    if mkdirs:
        dest.parent.mkdir(parents=True, exist_ok=True)
    try:
        source.rename(dest)
        return dest
    except OSError as e:
        # Check if the destination file already exists
        if e.errno == errno.EEXIST:
            if if_exists == 'fail':
                # User wants to receive this exception
                raise
            if if_exists == 'keep':
                # User wants to keep the destination file
                return dest
            # The alternative is to replace
            assert if_exists == 'replace', if_exists
            # User wants us to replace the file in the destination
            dest.unlink()
            return safe_move_file.sync(source, dest, mkdirs=mkdirs, if_exists=if_exists, cancel=cancel)
        if e.errno == errno.EXDEV:
            # Cross-device linking: We're trying to relink a file across a filesystem boundary. This is where the "safe" comes in.
            copy_file.sync(source, dest, if_exists=if_exists, mkdirs=True, preserve_symlinks=False, preserve_stat=True)
            source.unlink()
            return dest
        raise


def open_file_writer(fpath: Pathish, if_exists: IfExists = 'fail'):
    """
    Opens a native file as an `~IFileWriter`.
    """
    from ..storage import NativeFileStorage
    return NativeFileStorage('/').open_file_writer(fpath, if_exists=if_exists)


def read_file(fpath: Pathish) -> AsyncContextManager[AsyncIterable[bytes]]:
    """
    Opens a native file as an asyncronous byte iterable.
    """
    from ..storage import NativeFileStorage
    return NativeFileStorage('/').read_file(fpath)



from __future__ import annotations

import argparse
import enum
import re
import sys
import textwrap
from contextlib import nullcontext
from dataclasses import dataclass, field
from inspect import isclass
from typing import (Any, ContextManager, Iterable, Sequence, Type, cast, overload)

import dagon.tool.args
from typing_extensions import Final

from ..ext.base import BaseExtension
from ..util import T, U, Undefined, UndefinedType, first
from .option import NoSuchOptionError, Option

_NO_VALUE: Final = object()


class NoOptionValueError(ValueError):
    pass


class FulfilledOptions:
    """
    Type that holds the result of options that have been fulfilled.

    :param values: Iterable of pairs that define the values for each option.
        The value should match the type of the option, or be ``Undefined``, in
        which case asking for the option value will return ``None``
    """
    def __init__(self, values: Iterable[tuple[Option[Any], Any]]):
        self._values = dict(values)

    @overload
    def get(self, option: str) -> Any:
        ...

    @overload
    def get(self, option: str, *, default: Any) -> Any:
        ...

    @overload
    def get(self, option: Option[T]) -> T:
        ...

    @overload
    def get(self, option: Option[T], *, default: U) -> T | U:
        ...

    def get(self, option: Option[T] | str, default: Any = _NO_VALUE) -> T:
        """
        Get the value of the given option.

        If the option exists but no value was provided, returns ``None``. If the
        option does not exist, raises ``KeyError``.
        """
        if isinstance(option, str):
            found = first((o for o in self._values.keys() if o.name == option), default=None)
            if found is None:
                raise KeyError(f'No option with name "{option}"')
            option = found
        if not option in self._values:
            raise KeyError(f'No known option "{option.name}"')
        mine = self._values[option]
        if mine is _NO_VALUE:
            if default is not _NO_VALUE:
                return default
            raise NoOptionValueError(
                f'No value was provided for option "{option.name}", and no default value was specified')
        return cast(T, mine)


class OptionSet:
    """
    A collection of options.

    :param options: Initial options to populate with
    """
    def __init__(self, options: Iterable[Option[Any]] = ()):
        self._options: dict[str, Option[Any]] = {}
        for o in options:
            self.add(o)

    def add(self, opt: Option[T]) -> Option[T]:
        """
        Register the given option with this set. Returns the given option.

        Raises :class:`RuntimeError` if the given option name is already taken.
        """
        if opt.name in self._options:
            raise RuntimeError(f'Duplicate option "{opt.name}"')
        self._options[opt.name] = opt
        return opt

    def get(self, key: str) -> Option[Any] | None:
        """Get an option by name, or ``None`` if it does not exist in this set."""
        return self._options.get(key)

    @property
    def options(self) -> Iterable[Option[Any]]:
        """The options in this set"""
        return self._options.values()

    def _iter_fulfill(self, kvs: Iterable[str]) -> Iterable[tuple[Option[Any], Any]]:
        opt_re = re.compile(r'(.+?)=(.*)')
        all_keys = set(self._options.keys())
        for spec in kvs:
            mat = opt_re.match(spec)
            if not mat:
                raise ValueError(f'Invalid option specifier "{spec}" (should bey <key>=<value>)')
            key, val_str = mat.groups()
            opt = self.get(key)
            if opt is None:
                raise NoSuchOptionError(key)
            try:
                value = opt.parse_str(val_str)
            except Exception as e:
                raise ValueError(f'Failed to parse given string "{val_str}" as a value for option "{opt.name}"') from e
            err = opt.validate(value)
            if err is not None:
                raise RuntimeError(f'Invalid value for `{key}`: {err}')
            yield opt, value
            all_keys.remove(opt.name)

        for k in all_keys:
            opt = self._options[k]
            if opt.has_default:
                yield opt, opt.get_default()
            else:
                yield opt, _NO_VALUE

    def fulfill(self, kvs: Iterable[str]) -> FulfilledOptions:
        """
        Fulfill the options in this set using the list of strings, which should
        be of the format ``<key>=<value>``.

        If a given ``key`` does not correpsond to any option, raises
        ``NameError``. If an option was not given a value in ``kvs``, that
        option will return ``None`` in the returned :class:`FulfilledOptions`.
        """
        return FulfilledOptions(self._iter_fulfill(kvs))


class _OptionsArgs(dagon.tool.args.ParsedArgs):
    opts: Sequence[str] | None
    list_options: bool


@dataclass()
class _OptionsContext:
    options: OptionSet = field(default_factory=OptionSet)
    fulfilled: None | FulfilledOptions = None


def ctx_option_set() -> OptionSet:
    return _OptionsExt.app_data().options


def ctx_fulfilled_options() -> FulfilledOptions:
    c = _OptionsExt.app_data()
    if c.fulfilled is None:
        raise RuntimeError('Options have not yet been fulfilled')
    return c.fulfilled


def _format_docstr(s: str) -> Iterable[str]:
    paragraphs = textwrap.dedent(s).split('\n\n')
    for par in paragraphs:
        for line in textwrap.wrap(par):
            yield line.strip()
        yield ''  # Empty line after paragraph


def _option_help_annot(t: Type[Any] | UndefinedType) -> str:
    if t is Undefined:
        return '[no type]'
    assert isclass(t), t
    if issubclass(t, enum.Enum):
        return f'{t.__name__} {{' + ', '.join(str(f.value) for f in t) + '}'
    return t.__name__


def _list_options(opts: OptionSet, with_docs: bool = True) -> int:
    for opt in opts.options:
        print(f'{opt.name}: {_option_help_annot(opt.type)}')
        if with_docs and opt.doc:
            for line in _format_docstr(opt.doc):
                print(f'    {line}')
    return 0


class _OptionsExt(BaseExtension[_OptionsContext, None, None]):
    dagon_ext_name = 'dagon.options'

    def __init__(self) -> None:
        self._set = OptionSet()

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        grp = arg_parser.add_argument_group('Task Options')
        grp.add_argument('-o',
                         '--opt',
                         dest='opts',
                         metavar='<key>=<value>',
                         action='append',
                         help='Specify the value for a task option (can be repeated)')
        grp.add_argument('--list-options', '-lo', action='store_true', help='List available task options and exit')

    def handle_options(self, opts: argparse.Namespace) -> None:
        args = cast(_OptionsArgs, opts)
        if args.list_options:
            _list_options(ctx_option_set(), with_docs=not args.no_doc)
            sys.exit(0)
        self.app_data().fulfilled = ctx_option_set().fulfill(args.opts or ())

    def app_context(self) -> ContextManager[_OptionsContext]:
        return nullcontext(_OptionsContext())



"""
Abstract asynchronous filesystem-like data storage.
"""

from __future__ import annotations

import asyncio
import sqlite3
from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path, PosixPath, PurePosixPath
from typing import (IO, TYPE_CHECKING, AsyncContextManager, AsyncIterable, AsyncIterator, Awaitable, Iterable, Optional)

from typing_extensions import Protocol

from .. import fs, util
from .. import db as db_mod
from ..fs import IfExists, NPaths, Pathish, iter_pathish


class IFileWriter(Protocol):
    """
    An object that can be used to write data into storage.
    """
    def write_more(self, _data: bytes) -> Awaitable[None]:
        """
        Append the given data to the stored object.
        """
        ...


class IFileStorage(Protocol):
    """
    Access to filesystem-like storage.
    """
    def open_file_writer(self,
                         fpath: Pathish,
                         *,
                         if_exists: IfExists = 'fail') -> AsyncContextManager[Optional[IFileWriter]]:
        """
        Open a new :class:`~IFileWriter` to write a file into storage. The file
        is stored at the given 'fpath'. Parent "directories" are created
        implicitly.

        :param fpath: The path to open. Will be normalized to POSIX format.
        :param if_exists: The action to take if the given file already exists.

        This is a context manager. When the manager exits, the file is stored
        as-is and can no longer be appended to.

        If ``if_exists`` is `keep`, the context manager yields ``None`` when
        the file exists. Otherwise, the context manager always yields a valid
        :class:`~IFileWriter`.
        """
        ...

    def read_file(self, fpath: Pathish) -> AsyncContextManager[AsyncIterable[bytes]]:
        """
        Open the given file for reading. Returns an async-iterable of data chunks
        of the file.

        :param fpath: The path to open. Will be POSIX-normalized.

        :raises: :class:`FileNotFoundError` if 'fpath' does not exist.
        """
        ...


class _DatabaseFileWriter:
    "Writes a sequence of blobs into a SQLite database."

    def __init__(self, c: sqlite3.Cursor, file_id: int) -> None:
        self._cursor = c
        self._file_id = file_id
        self._counter = 0

    async def write_more(self, data: bytes) -> None:
        self._cursor.execute(
            r'''
            INSERT INTO dagon_storage_file_data (file_id, nth, data)
            VALUES (:file, :n, :data)
            ''',
            dict(file=self._file_id, n=self._counter, data=data),
        )
        self._counter += 1


_META_VERSION = 1


class _DatabaseFileStorage:
    def __init__(self, db: db_mod.Database, subkey: int) -> None:
        self._db = db
        self._subkey = subkey
        self._migrate_dbstore(db)

    @staticmethod
    def _do_migrate_dbstore(db: db_mod.Database) -> None:
        for s in [
                'DROP TABLE IF EXISTS dagon_storage_files',
                'DROP TABLE IF EXISTS dagon_storage_file_data',
                r'''
                CREATE TABLE dagon_storage_files (
                    file_id INTEGER PRIMARY KEY,
                    run_id NOT NULL REFERENCES dagon_runs ON DELETE CASCADE,
                    path TEXT NOT NULL,
                    UNIQUE(run_id, path)
                )''',
                r'''
                CREATE TABLE dagon_storage_file_data (
                    data_id INTEGER PRIMARY KEY,
                    file_id NOT NULL REFERENCES dagon_storage_files ON DELETE CASCADE,
                    nth INTEGER NOT NULL,
                    data BLOB NOT NULL,
                    UNIQUE(file_id, nth)
                )''',
        ]:
            db(s)

    @staticmethod
    def _migrate_dbstore(db: db_mod.Database) -> None:
        db.sqlite3_db.execute(r'''
            CREATE TABLE IF NOT EXISTS dagon_storage_meta (
                meta INTEGER NOT NULL
            )
        ''')
        meta = list(db('SELECT meta FROM dagon_storage_meta'))
        if meta and meta[0][0] == _META_VERSION:
            # Database schema is up-to-date
            return

        with util.recursive_transaction(db.sqlite3_db):
            _DatabaseFileStorage._do_migrate_dbstore(db)
            db('INSERT INTO dagon_storage_meta(meta) VALUES (:ver)', ver=_META_VERSION)

    @asynccontextmanager
    async def read_file(self, fpath: Pathish) -> AsyncIterator[AsyncIterable[bytes]]:
        """Implements IFileStorage.read_file()"""
        fpath = PurePosixPath(fpath)
        fid = list(
            self._db(
                r'''
                SELECT file_id
                  FROM dagon_storage_files
                 WHERE path = :path
                       AND run_id = :rid
                ''',
                path=str(fpath),
                rid=self._subkey,
            ))
        if not fid:
            raise FileNotFoundError(fpath)
        fid = fid[0][0]
        items: Iterable[tuple[bytes]] = self._db(
            r'''
            SELECT data
              FROM dagon_storage_file_data
             WHERE file_id = :fid
             ORDER BY nth
            ''',
            fid=fid,
        )
        yield self._read_items(items)

    async def _read_items(self, items: Iterable[tuple[bytes]]) -> AsyncIterable[bytes]:
        for i, in items:
            yield i

    def open_file_writer(self,
                         fpath: Pathish,
                         *,
                         if_exists: IfExists = 'fail') -> AsyncContextManager[Optional[IFileWriter]]:
        """
        Implements IFileStorage.open_file_writer
        """
        return self._open_file_writer(fpath, if_exists)

    @asynccontextmanager
    async def _open_file_writer(self, fpath: Pathish, if_exists: IfExists) -> AsyncIterator[Optional[IFileWriter]]:
        fpath = PurePosixPath(fpath)
        c = self._db.sqlite3_db.cursor()
        dml_cmd = {
            'fail': 'INSERT',
            'keep': 'INSERT',
            'replace': 'INSERT OR REPLACE',
        }[if_exists]
        if dml_cmd is None:
            raise ValueError(f'Invalid file storage `in_exists` `{if_exists}`')
        if if_exists == 'keep':
            exists: bool = util.first(
                self._db(r'''
                        SELECT count(*) FROM dagon_storage_files
                        WHERE path = :path AND run_id = :subkey
                        ''',
                         path=str(fpath),
                         subkey=self._subkey))[0]
            if exists:
                yield None
                return
        try:
            c.execute(
                fr'''
            {dml_cmd} INTO dagon_storage_files (run_id, path)
            VALUES (:run, :path)
            ''', dict(run=self._subkey, path=str(fpath)))
        except sqlite3.IntegrityError as e:
            if 'UNIQUE constraint failed' in str(e):
                raise FileExistsError(fpath) from e
            raise
        file_id = c.lastrowid
        fw: IFileWriter = _DatabaseFileWriter(c, file_id)
        yield fw


class _NativeFileWriter:
    def __init__(self, of: IO[bytes]):
        self._fd = of

    async def write_more(self, dat: bytes) -> None:
        remain = len(dat)
        while remain:
            remain -= self._fd.write(dat)


class NativeFileStorage:
    def __init__(self, directory: Pathish):
        self._path = Path(directory).resolve()

    @property
    def path(self) -> Path:
        return self._path

    def _normpath(self, given: Pathish) -> Path:
        p = PurePosixPath(given)
        if not p or p.parts[0] == '..':
            raise ValueError(f'Invalid path "{given}"')
        return Path(self.path / p)

    def open_file_writer(self,
                         fpath: Pathish,
                         *,
                         if_exists: IfExists = 'fail') -> AsyncContextManager[IFileWriter | None]:
        return self._open_file_writer(fpath, if_exists)

    @asynccontextmanager
    async def _open_file_writer(self, fpath: Pathish, if_exists: IfExists) -> AsyncIterator[IFileWriter | None]:
        fpath = self._normpath(fpath)

        fpath.parent.mkdir(exist_ok=True, parents=True)
        if if_exists == 'replace':
            with open(fpath, 'wb') as fd:
                yield _NativeFileWriter(fd)
                return
        try:
            fpath.parent.mkdir(exist_ok=True, parents=True)
            with open(fpath, 'xb') as fd:
                yield _NativeFileWriter(fd)
            return
        except FileExistsError:
            if if_exists == 'fail':
                raise
            if if_exists == 'keep':
                yield None
            else:
                assert 0

    @asynccontextmanager
    async def read_file(self, fpath: Pathish) -> AsyncIterator[AsyncIterable[bytes]]:
        fpath = self._normpath(fpath)
        with open(fpath, 'rb') as f:
            yield fs.read_blocks_from(f)


if TYPE_CHECKING:
    util.typecheck(IFileStorage)(_DatabaseFileStorage)
    util.typecheck(IFileWriter)(_DatabaseFileWriter)
    util.typecheck(IFileStorage)(NativeFileStorage)
    util.typecheck(IFileWriter)(_NativeFileWriter)


@asynccontextmanager
async def open_db_storage(*,
                          db: Optional[db_mod.Database] = None,
                          run_id: Optional[int] = None) -> AsyncIterator[IFileStorage]:
    if db is None or run_id is None:
        dbctx = db_mod.global_context_data()
        if not dbctx:
            raise RuntimeError('open_db_storage() requires a database argument or to be called within a task context')
        db = db or dbctx.database
        run_id = run_id if run_id is not None else dbctx.run_id
    async with db.transaction_context():
        yield _DatabaseFileStorage(db, run_id)


@asynccontextmanager
async def _ensure_storage(st: IFileStorage | None) -> AsyncIterator[IFileStorage]:
    if st is not None:
        yield st
        return

    dbctx = db_mod.global_context_data()
    if dbctx is None:
        raise RuntimeError(f'Invalid attempt to use database storage outside of a task context')
    db = dbctx.database
    async with db.transaction_context():
        yield _DatabaseFileStorage(db, dbctx.run_id)


async def store(files: NPaths,
                *,
                whence: Pathish | None = None,
                prefix: PurePosixPath | None = None,
                if_exists: IfExists = 'fail',
                into: IFileStorage | None = None) -> None:
    async with _ensure_storage(into) as into:
        whence = Path(whence or Path.cwd())
        prefix = PurePosixPath(prefix or '.')

        coros = (_store_file(into, f, whence, prefix, if_exists) for f in iter_pathish(files))
        await asyncio.gather(*coros)


async def _store_file(into: IFileStorage, fpath: Path, whence: Path, prefix: PurePosixPath,
                      if_exists: IfExists) -> None:
    fpath = fpath.resolve(strict=False)
    try:
        suffix = fpath.relative_to(whence).as_posix()
    except ValueError as e:
        raise RuntimeError(f'Cannot store file outside of working directory without a "whence" '
                           f'argument set to a parent directory of the file being stored. '
                           f'(filepath is "{fpath}", whence path is "{whence}")') from e
    stored_path = PosixPath(prefix.joinpath(suffix))
    await store_file_as(fpath, stored_path, into=into, if_exists=if_exists)


async def store_file_as(fpath: Pathish,
                        dest: Pathish,
                        *,
                        into: IFileStorage | None = None,
                        if_exists: IfExists = 'fail') -> None:
    async with AsyncExitStack() as stack:
        into = await stack.enter_async_context(_ensure_storage(into))
        writer = await stack.enter_async_context(into.open_file_writer(dest, if_exists=if_exists))
        if writer is None:
            assert if_exists == 'keep', if_exists
            return
        await store_file_in_writer(fpath, writer)


async def store_file_in_writer(fpath: Pathish, writer: IFileWriter) -> None:
    async for block in fs.read_blocks(fpath):
        await writer.write_more(block)


async def recover(fpath: Pathish, *, from_: IFileStorage | None = None) -> bytes:
    acc = b''
    async for part in recover_iter(fpath, from_=from_):
        acc += part
    return acc


async def recover_iter(fpath: Pathish, *, from_: IFileStorage | None = None) -> AsyncIterator[bytes]:
    async with AsyncExitStack() as st:
        from_ = await st.enter_async_context(_ensure_storage(from_))
        bufs = await st.enter_async_context(from_.read_file(fpath))
        async for b in bufs:
            yield b


if TYPE_CHECKING:
    f: IFileStorage = fs
    util.unused(f)



import asyncio
import datetime
from typing import Iterable
from ..event.cancel import CancellationToken
import sys

import pytest

from dagon import proc


def _join_out(out: Iterable[proc.ProcessOutputItem]) -> bytes:
    return b''.join(o.out for o in out)


@pytest.mark.asyncio
async def test_run_process_stdout():
    result = await proc.run([sys.executable, '-c', 'print("Hello!", end="")'])
    assert _join_out(result.output) == b'Hello!'


@pytest.mark.asyncio
async def test_run_proc_fail():
    result = await proc.run([sys.executable, '-c', 'import sys; sys.exit(42)'], check=False)
    assert result.retcode == 42
    assert _join_out(result.output) == b''


@pytest.mark.asyncio
async def test_run_cancel_via_token():
    token = CancellationToken()
    pipe = await proc.spawn([sys.executable, '-c', 'import time; time.sleep(5)'], cancel=token)
    token.cancel()
    with pytest.raises(asyncio.CancelledError):
        await pipe.result


@pytest.mark.asyncio
async def test_run_timeout():
    with pytest.raises(TimeoutError):
        await proc.run([sys.executable, '-c', 'import time; time.sleep(1)'], timeout=datetime.timedelta(milliseconds=5))


@pytest.mark.asyncio
async def test_run_no_timeout():
    await proc.run([sys.executable, '-c', 'import time; time.sleep(0.2)'], timeout=datetime.timedelta(seconds=1))


@pytest.mark.asyncio
async def test_run_empty_stdin():
    result = await proc.run([sys.executable, '-c', 'import sys; s = sys.stdin.read(); print(s)'],
                            timeout=datetime.timedelta(seconds=1),
                            check=False)
    assert result.retcode == 0
    assert _join_out(result.output) == proc.OS_PREFERRED_NEWLINE


@pytest.mark.asyncio
async def test_run_simple_stdin():
    result = await proc.run(
        [sys.executable, '-c', 'import sys; a = input("a:"); b = input("b:"); print("got:", a, b, sys.stdin.read())'],
        stdin='eggs\nbacon\n',
        timeout=datetime.timedelta(seconds=1),
        check=False)
    assert result.retcode == 0
    assert _join_out(result.output) == b'a:b:got: eggs bacon ' + proc.OS_PREFERRED_NEWLINE


@pytest.mark.asyncio
async def test_output_handler():
    acc = b''

    def handle(data: proc.ProcessOutputItem):
        nonlocal acc
        acc += data.out

    res = await proc.run([sys.executable, '-c', 'import sys; sys.stdout.write("abc\\n")'], on_output=handle)
    assert _join_out(res.output) == b'abc' + proc.OS_PREFERRED_NEWLINE
    assert acc == b'abc' + proc.OS_PREFERRED_NEWLINE

    acc = b''

    res = await proc.run([sys.executable, '-c', 'import sys; sys.stdout.write("abc")'], on_output=handle)
    assert _join_out(res.output) == b'abc'
    assert acc == b'abc'



import enum
from pathlib import Path

import pytest

from . import option as mod


def test_simple_option() -> None:
    mod.Option('foo', type=int, parse=int)


def test_bool() -> None:
    assert mod.convert_bool('1')
    assert mod.convert_bool('y')
    assert mod.convert_bool('yes')
    assert mod.convert_bool('true')
    assert not mod.convert_bool('0')
    assert not mod.convert_bool('n')
    assert not mod.convert_bool('no')
    assert not mod.convert_bool('false')


class MyEnum(enum.Enum):
    Value = 'value'
    Cat = 'cat'


def test_valid_types() -> None:
    assert mod.is_valid_option_type(int)
    assert mod.is_valid_option_type(float)
    assert mod.is_valid_option_type(str)
    assert mod.is_valid_option_type(Path)
    assert mod.is_valid_option_type(MyEnum)
    assert not mod.is_valid_option_type(tuple)


def test_parse_option() -> None:
    assert mod.parse_value_str(MyEnum, 'cat') == MyEnum.Cat
    assert mod.parse_value_str(int, '11') == 11
    with pytest.raises(ValueError):
        mod.parse_value_str(MyEnum, 'invalid')



from __future__ import annotations

import sys
from contextlib import ExitStack, asynccontextmanager, contextmanager
from typing import AsyncIterator, Iterator, TextIO

from dagon.event import Event, events
from dagon.task.dag import OpaqueTask
from dagon.util import AsyncNullContext, unused

from .base import BaseExtension
from .iface import OpaqueTaskGraphView


class PrintCollector(BaseExtension[None, 'PrintCollector', None]):
    dagon_ext_name = 'dagon.print'
    dagon_ext_requires = ['dagon.events']

    def __init__(self):
        self._stdout: TextIO | None = None
        self._stderr: TextIO | None = None
        self.on_print = Event[str]()

    @contextmanager
    def _patch_stdout(self) -> Iterator[TextIO]:
        old_stdout = sys.stdout
        old_stderr = sys.stderr

        evh = self

        class StdIOPatch:
            def write(self, dat: str) -> None:  # pylint: disable=no-self-use
                # pylint: disable=protected-access
                evh.on_print.emit(dat)
                sys.__stdout__.write(dat)

            def flush(self) -> None:
                pass

        sys.stderr = sys.stdout = StdIOPatch()  # type: ignore
        try:
            self._stdout = old_stdout  # type: ignore
            yield old_stdout
        finally:
            self._stdout = None
            sys.stdout = old_stdout
            sys.stderr = old_stderr

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[PrintCollector]:
        unused(graph)
        with ExitStack() as st:
            st.enter_context(self._patch_stdout())
            yield self

    def task_context(self, task: OpaqueTask):
        events.register('dagon.print', self.on_print)
        return AsyncNullContext()



import sqlite3
from datetime import datetime
from pathlib import Path

import pytest

from .. import db as mod


def get_mem_db() -> mod.Database:
    return mod.Database.get_or_create(':memory:')


# Mark tests as asyncio to ensure there is a thread loop for the database
@pytest.mark.asyncio
def test_create() -> None:
    mod.get_ready_sqlite_db(':memory:')


@pytest.mark.asyncio
def test_get_existing() -> None:
    test_db_s = 'foo.db'
    test_db = Path(test_db_s)
    if test_db.exists():
        test_db.unlink()
    with pytest.raises(mod.MissingDatabase):
        mod.get_existing_sqlite_db(test_db)

    conn = sqlite3.connect(test_db_s)
    conn.commit()
    with pytest.raises(mod.InvalidDatabase):
        mod.get_existing_sqlite_db(test_db)

    conn.execute('create table dagon_meta (hash)')
    phony_version = 'bad'
    conn.execute(f'INSERT INTO dagon_meta VALUES (\'{phony_version}\')')
    conn.commit()

    try:
        mod.get_existing_sqlite_db('foo.db')
        assert False, 'Opening bad database did not raise'
    except mod.IncorrectDatabaseVersion as e:
        assert e.path == test_db
        assert e.version == phony_version
        assert e.expected_version == mod.schema_hash()


@pytest.mark.asyncio
def test_new_run_id() -> None:
    db = get_mem_db()
    run_id = db.new_run_id()
    assert run_id == 1
    run_id_2 = db.new_run_id()
    assert run_id_2 == 2


@pytest.mark.asyncio
def test_persist() -> None:
    db = get_mem_db()
    assert db.load_persist('foo', None) is mod.Undefined
    db.set_persist('foo', None, 12)
    assert db.load_persist('foo', None) == 12


@pytest.mark.asyncio
def test_task_events() -> None:
    db = get_mem_db()
    run = db.new_run_id()
    t = db.get_task_rowid('test')
    trun_id = db.add_task_run(run=run, task=t)
    db.store_task_event(trun_id, 'start')


@pytest.mark.asyncio
def test_store_proc_exec() -> None:
    db = get_mem_db()
    run = db.new_run_id()
    t = db.get_task_rowid('test')
    trun_id = db.add_task_run(run=run, task=t)
    db.store_proc_execution(
        trun_id,
        cmd=['foo', 'bar'],
        cwd=Path('/nowhere'),
        output=(),
        retc=12,
        start_time=datetime.now(),
        duration=12,
    )



"""
Module ``dagon.util``
#####################

General utilities
"""
from __future__ import annotations

import enum
import contextvars
import difflib
import sqlite3
import types
from contextlib import ExitStack, contextmanager
from typing import (TYPE_CHECKING, Any, AsyncContextManager, Awaitable, Callable, ContextManager, Generator, Generic,
                    Iterable, Iterator, Type, TypeVar, cast, overload)

from typing_extensions import Protocol

T = TypeVar('T')
"A generic invariant type variable"
T_co = TypeVar('T_co', covariant=True)
"A generic covariant type variable"
U = TypeVar('U')
"A second generic invariant type variable"


class UndefinedType(enum.Enum):
    """
    The type of the generic :const:`Undefined` constant.
    """
    Undefined = 0

    def __repr__(self) -> str:
        return 'Undefined'


Undefined = UndefinedType.Undefined
"""
An 'undefined' constant to be used to represent the absence of parameter/return
values where `None` is within the domain of that parameter or return value.
"""


@overload
def first(items: Iterable[T]) -> T:
    ...


@overload
def first(items: Iterable[T], *, default: U) -> T | U:
    ...


def first(items: Iterable[T], **kw: U) -> T | U:
    """
    Obtain the first element of an iterable

    :param items: An iterable object
    :param default: A default value to return in case of an empty iterable.

    :raises ValueError: If `items` is empty and no `default` was provided
        omitted, `ValueError` will be raised in case of an empty iterable.
    """
    for n in items:
        return n
    if 'default' not in kw:
        raise ValueError(f'No first item in an empty iterable ({items!r})')
    return kw['default']


def unused(*args: Any) -> None:
    """Does nothing. Used to mark the given arguments as unused."""
    args  # pylint: disable=pointless-statement


class NoneSuch(Generic[T]):
    """
    Represents a missing value of type ``T``, and may have a candidate.

    :param key: The key that was attempted to look up.
    :param candidate: The closest match to ``key``.
    """
    def __init__(self, key: str, candidate: None | T) -> None:
        self._key = key
        self._cand = candidate

    @property
    def key(self) -> str:
        """The key that was invalid"""
        return self._key

    @property
    def candidate(self) -> T | None:
        """The best-matching candidate"""
        return self._cand

    def __repr__(self) -> str:
        return f'<NoneSuch given="{self.key}" nearest={self.candidate!r}>'


class Opaque(Protocol):
    """
    An opaque type. Unlike :class:`Any`, does not implicitly cast anywhere.
    """
    def ___opaque___(self) -> None:
        ...


def dot_kebab_name(name: str) -> str:
    """
    Convert a ``snake_case`` name into a ``dot.kebab-case`` name, and strip
    leading/trailing underscores. Double underscores ``__`` are converted to
    a dot ``.``.

    :param name: The name to convert.

    .. list-table:: Converted Names
        :widths: auto
        :header-rows: 1

        * - Input
          - Result
        * - ``meow``
          - ``meow``
        * - ``john_doe``
          - ``john-doe``
        * - ``do_thing_``
          - ``do-thing``
        * - ``_do_thing``
          - ``do-thing``
        * - ``foo__bar``
          - ``foo.bar``
        * - ``foo__bar_baz``
          - ``foo.bar-baz``
    """
    return name.replace('__', '.').strip('_').replace('_', '-')


def on_context_exit(cb: Callable[[], None]) -> ContextManager[None]:
    """Create a context manager that simply calls the given callback on ``__exit__``"""
    st = ExitStack()
    st.callback(cb)
    return cast(ContextManager[None], st)


def scope_set_contextvar(cvar: contextvars.ContextVar[T], value: T) -> ContextManager[None]:
    """
    Create a context manager that sets the
    `context variable <contextvars.ContextVar>` to the given value, then resets
    the value on ``__exit__``
    """
    tok = cvar.set(value)
    return on_context_exit(lambda: cvar.reset(tok))


class ReadyAwaitable(Generic[T]):
    """
    An `Awaitable` object that when awaited will immediately resolve to a given
    value without suspending the awaiting coroutine.

    :param value: The value that will be returned from the ``await`` expression.
    """
    def __init__(self, value: T):
        self._value = value

    def __await__(self) -> Generator[None, None, T]:
        return self._value
        # Unreachable 'yield', but makes this function into a generator
        yield None


class AsyncNullContext(Generic[T]):
    """
    Like `~contextlib.nullcontext`, but implements an `~typing.AsyncContextManager`
    """
    def __init__(self, value: T = None) -> None:
        self._value = value

    def __aenter__(self) -> Awaitable[T]:
        return ReadyAwaitable(self._value)

    def __aexit__(self, _exc_t: Type[BaseException] | None, _exc: BaseException | None,
                  _tb: types.TracebackType | None) -> Awaitable[None]:
        return ReadyAwaitable(None)


def typecheck(iface: Type[T]) -> Callable[[Type[T]], None]:
    """
    Given a type, return a callable that accepts that type. This can be used
    to insert type checks into modules. Should not be called at runtime: guard
    this with a `typing.TYPE_CHECKING` condition.
    """
    unused(iface)
    assert False, TypeError('typecheck() should never by called at runtime')
    return lambda f: None  # Unreachable, but makes Pylint happy


if TYPE_CHECKING:
    typecheck(AsyncContextManager[None])(AsyncNullContext[None])
    typecheck(AsyncContextManager[int])(AsyncNullContext[int])


def nearest_matching(given: str, of: Iterable[T], key: Callable[[T], str]) -> T | None:
    """
    Find the object that is "nearest" to the `given` string based on the string
    distance. Each object should be mapped to a string with the `key` function.

    If `of` is empty, returns `None`.
    """
    return max(of, key=lambda t: difflib.SequenceMatcher(None, given, key(t)).ratio(), default=None)


def fixup_dataclass_docs(cls: Type[T]) -> Type[T]:
    cls.__init__.__qualname__ = f'{cls.__name__}.__init__'
    return cls


def ensure_awaitable(val: Awaitable[T] | T) -> Awaitable[T]:
    """
    Take an object that may or may not be `~typing.Awaitable` at runtime, and
    ensure that it is awaitable. If the given object is not awaitable, it will
    be wrapped in a `.ReadyAwaitable`.
    """
    if isinstance(val, Awaitable):
        return cast(Awaitable[T], val)
    return ReadyAwaitable(val)


@contextmanager
def recursive_transaction(db: sqlite3.Connection) -> Iterator[None]:
    """
    Create a scoped "recursive" transaction on the given SQLite database
    connection.

    If the database is already in a transaction, this context manager is a no-op
    on entry and exit. Otherwise, creates a transaction that is COMMITed on
    exit, or ROLLedBACK if the scope exists via exception.

    asserts that no one closed the transaction outside of our watch.
    """
    if db.in_transaction:
        try:
            yield
        finally:
            assert db.in_transaction, 'transaction was ended prematurely'
        return
    db.execute('BEGIN')
    try:
        yield
    except:
        assert db.in_transaction, 'transaction was ended prematurely'
        db.execute('ROLLBACK')
        raise
    else:
        assert db.in_transaction, 'transaction was ended prematurely'
        db.execute('COMMIT')



import argparse
from typing import (AsyncContextManager, Awaitable, ContextManager, Generic, Iterable, TypeVar)

from dagon.core import ll_dag
from dagon.core.result import NodeResult
from dagon.task.dag import OpaqueTask
from typing_extensions import Protocol, runtime_checkable

AppDataT = TypeVar('AppDataT', covariant=True)
GlobalDataT = TypeVar('GlobalDataT', covariant=True)
TaskDataT = TypeVar('TaskDataT', covariant=True)

OpaqueTaskGraphView = ll_dag.DAGView[OpaqueTask]


@runtime_checkable
class IExtension(Generic[AppDataT, GlobalDataT, TaskDataT], Protocol):
    """
    The minimal interface required for any Dagon extension.
    """
    @property
    def dagon_ext_requires(self) -> Iterable[str]:
        """Extensions required by this extension"""
        ...

    @property
    def dagon_ext_requires_opt(self) -> Iterable[str]:
        """Extensions that are optionally used by this extension"""
        ...

    @property
    def dagon_ext_name(self) -> str:
        """The name of the extension"""
        ...

    def app_context(self) -> ContextManager[AppDataT]:
        """Synchronous application-level context setup for the extension."""
        ...

    def add_options(self, arg_parser: argparse.ArgumentParser) -> None:
        """Augment the command-line argument parser for this extension"""
        ...

    def handle_options(self, opts: argparse.Namespace) -> None:
        """Handle the commandd-line arguments for this extension"""
        ...

    def global_context(self, graph: OpaqueTaskGraphView) -> AsyncContextManager[GlobalDataT]:
        """The DAG-global context for this extension. Invoked during DAG execution startup/teardown."""
        ...

    def task_context(self, task: OpaqueTask) -> AsyncContextManager[TaskDataT]:
        """The task-local context for this extension. Invoked during task startup/teardown."""
        ...

    def notify_result(self, result: NodeResult[OpaqueTask]) -> Awaitable[None]:
        """Called by the DAG executor when a task completes, notifying the extension of the task's result."""
        ...



"""
Module for drawing a dependency graph
"""

import tempfile

from dagon.db import Database, RunID

try:
    import graphviz
    HAVE_GRAPHVIZ = True
except ModuleNotFoundError as e:
    if e.name != 'graphviz':
        # pyright: reportConstantRedefinition=false
        HAVE_GRAPHVIZ = False
        raise
    # pyright: reportConstantRedefinition=false
    HAVE_GRAPHVIZ = False


def generate_graph(db: Database, run: RunID) -> str:
    tmpdir = tempfile.mkdtemp()
    dig = graphviz.Digraph(comment=f'Graph generated by Dagon for run {run}', directory=tmpdir)
    dig.attr(rankdir='LR', splines='spline')

    tasks = db(
        r'''
        SELECT DISTINCT dependent, end_state
          FROM dagon_task_deps_rel
    INNER JOIN dagon_tasks
            ON name = dependent
  NATURAL LEFT OUTER JOIN dagon_task_runs
         WHERE run_id=:r
        ''',
        r=run,
    )
    for tname, end_state in tasks:
        col = {
            'pending': 'cyan',
            'running': 'blue',
            'failed': 'red',
            'succeeded': 'green',
            'cancelled': 'orange',
            None: 'gray',  # No dagon_task_runs row for this target
        }[end_state]
        dig.node(tname, color=col, fontname='mono', penwidth='5', shape='rect')

    rows = db(r'''
        SELECT dep.dependent as target,
               dep.depends_on as dep,
               0
          FROM dagon_task_deps_rel dep
          JOIN dagon_tasks t
            ON t.name = dep.dependent
         WHERE run_id=:rid
      ORDER BY name
    ''',
              rid=run)

    for row in rows:
        out, in_, oo = row
        if in_ is None:
            continue
        style = 'dashed' if oo else 'solid'
        dig.edge(in_, out, style=style)

    dig.render(format='png', view=True)
    return str(dig)



"""
Module ``dagon.option``
#######################

Create run-time execution options to customize task behaviors.

Task can define customization options using this API to tweak task behaviors.
The value assigned to any option is fixed at task graph invocation and is
uniform across a task graph execution.

Using `.option.add` to create new options while declaring tasks, and use
`.option.Option.get` or `.option.value_of` to obtain the value provided to an
option at runtime.

.. data:: OptionType
    :canonical: dagon.option.option.OptionType
    :type: str | bool | pathlib.Path | int | float | enum.Enum

    A type union of any type that can be used with `.Option`.

    .. _option-types:

    Dagon supports several types by default:

    - `str` for strings.
    - `bool` for true/false values.
    - :class:`pathlib.Path` to represent filesystem paths.
    - `int` for integral values.
    - `float` for floating point numbers.
    - Any subclass of :class:`enum.Enum`. It is recommended to use `str` keys
      with enums when using such types.

    In addition, Dagon supports any class that presents a `__dagon_parse_opt__`
    static method, which should receive a string and return an instance of the
    class.

.. data:: OptionT
    :canonical: dagon.option.option.OptionT

    A type variable bound to `.OptionType`.
"""

from __future__ import annotations

from typing import Any, Callable, Type, TypeVar, cast, overload

from ..util import T, U
from .ext import ctx_fulfilled_options, ctx_option_set
from .option import Option, SimpleOptionType, get_type_parser

__all__ = ['Option', 'add', 'value_of']

SimpleT = TypeVar('SimpleT', bound=SimpleOptionType)


@overload
def add(name: str,
        type: Type[SimpleT],
        *,
        doc: str = '',
        validate: Callable[[SimpleT], str | None] | None = None,
        default: U | Callable[[str], U] = ...) -> Option[SimpleT | U]:
    ...


@overload
def add(name: str,
        type: Type[T] = ...,
        *,
        doc: str = '',
        parse: Callable[[str], T],
        default: U | Callable[[str], U] = ...) -> Option[T | U]:
    ...


def add(name: str,
        type: Type[T] | None = None,
        *,
        parse: Callable[[str], T] | None = None,
        doc: str = '',
        validate: Callable[[Any], str | None] | None = None,
        **kwargs: Any) -> Option[Any]:
    """
    Add an option to the current task graph with the given name and type.
    """
    oset = ctx_option_set()
    if type is not None:
        parse = get_type_parser(cast(Type[SimpleOptionType], type))
    assert parse, 'option.add requires a "parse" function or a supported "type"'
    if 'default' not in kwargs:
        return oset.add(Option[Any](name, type=type, parse=parse, doc=doc, validate=validate))
    d: Any = kwargs.pop('default')
    if callable(d):
        return oset.add(Option[Any](name, type=type, parse=parse, doc=doc, validate=validate, calc_default=d))
    else:
        return oset.add(Option[Any](name, type=type, parse=parse, doc=doc, validate=validate, default=d))


@overload
def value_of(opt: Option[T], *, default: U = ...) -> T | U:
    ...


@overload
def value_of(opt: str, *, default: Any = ...) -> Any:
    ...


def value_of(opt: Option[T] | str, **kw: Any) -> T | None:
    """Obtain the value of the specified option"""
    return ctx_fulfilled_options().get(opt, **kw)



"""
Module ``dagon.event``
######################

Event handling

.. data:: events
    :type: dagon.event.EventMap

    A task-local `event map <dagon.event.EventMap>`.

    Comes pre-loaded with events:

    ``dagon.interval-start``
        `Event[str] <.Event>` : An event fired by `.interval_start`

    ``dagon.interval-end``
        `Event[None] <.Event>` : An event fired by `.interval_end`

    ``dagon.mark``
        `Event[str] <.Event>` : An event fired by `.mark`

    .. note:: Using this event map is only allowed following the initialization
        of the ``dagon.events`` extension, and only within the context of a
        task's execution. Each task receives its own fresh event map object.
"""

from __future__ import annotations

from contextlib import asynccontextmanager, contextmanager
import contextvars
from typing import Any, AsyncIterator, Callable, Iterator

from ..ext.base import BaseExtension
from ..ext.iface import OpaqueTaskGraphView
from ..task.dag import OpaqueTask
from ..util import T, scope_set_contextvar, unused
from .cancel import CancellationToken, CancelLevel, raise_if_cancelled
from .event import ConnectionToken, Event, EventMap

__all__ = [
    'events',
    'Event',
    'CancellationToken',
    'raise_if_cancelled',
    'ConnectionToken',
    'CancelLevel',
    'EventMap',
    'interval_context',
    'interval_end',
    'interval_start',
    'mark',
]

_DEFAULT_EV_MAP = EventMap()

_EVENTS_CTX = contextvars.ContextVar[EventMap]('_EVENTS_CTX', default=_DEFAULT_EV_MAP)


class _EventsExt(BaseExtension[None, EventMap, EventMap]):
    dagon_ext_name = 'dagon.events'

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[EventMap]:
        prev = _EVENTS_CTX.get()
        map = prev.child()
        with scope_set_contextvar(_EVENTS_CTX, map):
            with CancellationToken.ensure_context_local():
                yield map

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[EventMap]:
        glb = self.global_data()
        child = glb.child()
        with scope_set_contextvar(_EVENTS_CTX, child):
            child.register('dagon.interval-start', Event[str]())
            child.register('dagon.interval-end', Event[None]())
            child.register('dagon.mark', Event[None]())
            yield child


class _EventsContextLookup:
    @staticmethod
    def _ctx() -> EventMap:
        return _EVENTS_CTX.get()

    def __getitem__(self, key: str) -> Event[Any]:
        return self._ctx()[key]

    def get(self, key: str) -> Event[Any] | None:
        return self._ctx().get(key)

    def get_or_register(self, name: str, factory: Callable[[], Event[T]] = Event[T]) -> Event[T]:
        return self._ctx().get_or_register(name, factory)

    def register(self, name: str, ev: Event[T]) -> Event[T]:
        return self._ctx().register(name, ev)


events = _EventsContextLookup()


def interval_start(name: str) -> None:
    """
    Fire a ``dagon.interval-start`` event.

    .. note:: May only be called within a task-executing context.
    """
    events['dagon.interval-start'].emit(name)


def interval_end() -> None:
    """
    Fire a ``dagon.interval-end`` event.

    .. note:: May only be called within a task-executing context.
    """
    events['dagon.interval-end'].emit(None)


@contextmanager
def interval_context(name: str) -> Iterator[None]:
    """
    Create a scope for an `.interval_start` and `.interval_end` block with
    the given name.

    .. note:: May only be called within a task-executing context.
    """
    interval_start(name)
    try:
        yield
    finally:
        interval_end()


def mark(name: str) -> None:
    """
    Emit a ``dagon.mark`` event with the given name.

    .. note:: May only be called within a task-executing context.
    """
    events['dagon.mark'].emit(name)


unused(_EventsExt)



from __future__ import annotations

import asyncio
import contextvars
import itertools
import os
import signal
from contextlib import AsyncExitStack, asynccontextmanager, contextmanager
from types import FrameType
from typing import (TYPE_CHECKING, Any, AsyncIterator, Awaitable, Callable, Iterable, Iterator, Mapping, NamedTuple,
                    Sequence, TypeVar, cast, overload)

from dagon.core.result import Failure

from ..core import LowLevelDAG, NodeResult, Success, exec
from ..util import (NoneSuch, Opaque, first, nearest_matching, scope_set_contextvar)
from .task import DisabledTaskError, InvalidTask, Task

if TYPE_CHECKING:
    from dagon.event.cancel import CancellationToken

T = TypeVar('T')

OpaqueTask = Task[Opaque]


class NoDependencyError(RuntimeError):
    def __init__(self, from_task: Task[Any], to_task: Task[Any]):
        super().__init__(f'Task "{from_task.name}" does not have a direct dependency on "{to_task.name}"')
        self.from_task = from_task
        self.to_task = to_task


class IsOrderOnlyDependencyError(RuntimeError):
    def __init__(self, from_task: Task[Any], to_task: Task[Any]) -> None:
        super().__init__(f'The dependcy of "{from_task.name}" on "{to_task.name}" is an order-only dependency')
        self.from_task = from_task
        self.to_task = to_task


async def default_execute(task: Task[T]) -> T:
    return await task.function()


class TaskDAG:
    def __init__(self, name: str) -> None:
        self._name = name
        self._tasks: dict[str, OpaqueTask] = {}

    @property
    def tasks(self) -> Iterable[Task[Any]]:
        """Tasks associated with this DAG"""
        return self._tasks.values()

    def add_task(self, task: Task[T]) -> Task[T]:
        if task.name in self._tasks:
            raise RuntimeError(f'More than one task with name "{task.name}"')
        for dep in task.depends:
            if dep.dep_name not in self._tasks.keys():
                # pylint: disable=cell-var-from-loop
                cand = nearest_matching(dep.dep_name, self.tasks, key=lambda t: t.name)
                raise InvalidTask(NoneSuch[OpaqueTask](dep.dep_name, cand))
        self._tasks[task.name] = cast(OpaqueTask, task)
        return task

    def _collect(self, mark: str, stack: list[str], marked: set[str]) -> Iterable[OpaqueTask]:
        if mark in marked:
            return

        # Check that we actually have this task
        if mark not in self._tasks:
            cand = nearest_matching(mark, self.tasks, key=lambda t: t.name)
            raise InvalidTask(NoneSuch[OpaqueTask](mark, cand))

        # Append to the search stack
        stack.append(mark)
        t = self._tasks[mark]

        # Check if it was disabled:
        if t.is_disabled:
            assert t.disabled_reason
            raise DisabledTaskError(stack, t.disabled_reason)

        # Yield this item
        yield t
        marked.add(t.name)

        # Yield each hard dependency
        for d in t.depends:
            if d.is_order_only or d.dep_name in marked:
                continue
            yield from self._collect(d.dep_name, stack, marked)

        # Pop from the search stack
        stack.pop()

    def low_level_graph(self, marks: Iterable[str]) -> LowLevelDAG[Task[Any]]:
        by_name = {t.name: t for t in (itertools.chain.from_iterable(self._collect(m, [], set()) for m in marks))}
        return LowLevelDAG[Task[Any]](
            nodes=by_name.values(),
            edges=itertools.chain.from_iterable(
                ((by_name[d.dep_name], t) for d in t.depends if d.dep_name in by_name) for t in by_name.values()),
        )

    def create_executor(self,
                        marks: Iterable[str],
                        *,
                        exec_fn: Callable[[Task[T]], Awaitable[T]] = default_execute) -> exec.SimpleExecutor[Task[Any]]:
        graph = self.low_level_graph(marks)
        return TaskExecutor[Opaque](graph, exec_fn=exec_fn, catch_signals=True, fail_cancels=True)

    async def execute(
            self,
            marks: Iterable[str],
            *,
            exec_fn: Callable[[Task[T]], Awaitable[T]] = default_execute) -> Mapping[Task[Any], NodeResult[Task[Any]]]:
        ex = self.create_executor(marks, exec_fn=exec_fn)
        return await ex.run_all()


class _ExecCtx(NamedTuple):
    exe: TaskExecutor[Any]
    task: Task[Any]


_CTX_EXEC = contextvars.ContextVar[_ExecCtx]('_CTX_EXECUTING_GRAPH')


def current_task() -> Task[Any]:
    return _CTX_EXEC.get().task


@overload
def result_of(task: Task[T]) -> Awaitable[T]:
    ...


@overload
def result_of(task: str) -> Awaitable[Any]:
    ...


def result_of(task: Task[Any] | str) -> Awaitable[Any]:
    """
    Obtain the result of the given task.

    The other task must be a direct dependency of the currently executing task,
    and must not be an order-only dependency.
    """
    ctx = _CTX_EXEC.get()
    ctx_task = ctx.task
    if isinstance(task, str):
        find = (n for n in ctx.exe.graph.all_nodes if n.name == task)
        try:
            task, = find
        except ValueError:
            raise NameError(f'No known task named "{task}"')
    find = (d for d in ctx_task.depends if d.dep_name == task.name)
    try:
        dep = first(find)
    except ValueError:
        raise NoDependencyError(ctx_task, task)
    if dep.is_order_only:
        raise IsOrderOnlyDependencyError(ctx_task, task)
    ares = ctx.exe.result_of(task)
    return _await_result(ares, task.name)


async def _await_result(ares: Awaitable[NodeResult[Any]], name: str) -> Any:
    res = await ares
    if isinstance(res.result, Success):
        return res.result.value
    raise RuntimeError(f'Attempted to obtain result of task "{name}", which is failed/cancelled')


class TaskExecutor(exec.SimpleExecutor[Task[T]]):
    """
    A DAG executor that executes task objects and allows the results of tasks
    to be shared between each other using the `.result_of` function.
    """
    def __init__(self, graph: LowLevelDAG[Task[T]], exec_fn: Callable[[Task[T]], Awaitable[T]], *, catch_signals: bool,
                 fail_cancels: bool):
        super().__init__(graph, exec_fn)
        self.__graph = graph
        self.__catch_signals = catch_signals
        self.__fail_cancels = fail_cancels

    @property
    def graph(self):
        """The low-level graph executed by this executor"""
        return self.__graph

    @asynccontextmanager
    async def running_context(self) -> AsyncIterator[None]:
        from dagon.event.cancel import CancellationToken
        async with AsyncExitStack() as st:
            await st.enter_async_context(super().running_context())
            tok = CancellationToken.get_context_local()
            if tok is None:
                tok = CancellationToken()
                tok.set_context_local(tok)
                st.callback(lambda: tok.set_context_local(None))
            handle_signals: list[int] = []
            if self.__catch_signals:
                handle_signals.extend((signal.SIGINT, signal.SIGTERM))
                if os.name == 'nt':
                    handle_signals.append(signal.SIGBREAK)
                else:
                    handle_signals.append(signal.SIGQUIT)
            st.enter_context(self._handle_signals(handle_signals, tok))
            yield

    @contextmanager
    def _handle_signals(self, sigs: Sequence[int], cancel: CancellationToken) -> Iterator[None]:
        loop = asyncio.get_event_loop()
        prev_handlers = [signal.signal(n, lambda n, fr: self._do_stop_threadsafe(n, fr, cancel, loop)) for n in sigs]
        try:
            yield
        finally:
            for sig, prev in zip(sigs, prev_handlers):
                signal.signal(sig, prev)

    @staticmethod
    def _do_stop_threadsafe(_signum: int, _frame: None | FrameType, cancel: CancellationToken,
                            loop: asyncio.AbstractEventLoop) -> None:
        loop.call_soon_threadsafe(cancel.cancel)

    async def do_run_node(self, node: Task[T]) -> NodeResult[Task[T]]:
        """
        Overridden from `SimpleExecutor.do_run_node`. Sets up the context to
        allow result sharing between tasks. Calls `.do_run_task` to execute
        the task within the scope that allows for result sharing.
        """
        from ..event import CancellationToken
        with scope_set_contextvar(_CTX_EXEC, _ExecCtx(self, node)):
            result = await self.do_run_task(node)
            if isinstance(result.result, Failure) and self.__fail_cancels:
                cancel = CancellationToken.get_context_local()
                if cancel:
                    cancel.cancel()
            return result

    async def do_run_task(self, node: Task[T]) -> NodeResult[Task[T]]:
        return await super().do_run_node(node)


_CURRENT_DAG = contextvars.ContextVar['TaskDAG | None']('_CURRENT_DAG', default=None)


def current_dag() -> TaskDAG:
    """
    Get the current thread/task-local Dag instance.
    """
    dag = _CURRENT_DAG.get()
    if dag is None:
        raise RuntimeError('There is no dag set for the current thread')
    return dag


def set_current_dag(dag: TaskDAG | None) -> None:
    """
    Set the current thread-local Dag instance.
    """
    _CURRENT_DAG.set(dag)


@contextmanager
def populate_dag_context(dag: TaskDAG) -> Iterator[None]:
    """
    Set the current thread-local Dag to ``dag``, and then restore the old
    thread-local Dag value when the context manager exists.

    :rtype: ContextManager[None]
    """
    tok = _CURRENT_DAG.set(dag)
    try:
        yield
    finally:
        _CURRENT_DAG.reset(tok)



"""
Module for creating/managing persistent data.
"""

from contextlib import asynccontextmanager, contextmanager
import contextvars
from dagon.ext.iface import OpaqueTaskGraphView
from dagon.task.dag import OpaqueTask
from dagon.util import typecheck, unused
from typing import TYPE_CHECKING, Any, AsyncIterator, Iterator, NamedTuple, Optional, overload
from typing_extensions import Protocol

from ..ext.base import BaseExtension
from .. import db


class IPersistentDataAccess(Protocol):
    def set(self, key: str, value: Any) -> None:
        ...

    @overload
    def get(self, key: str) -> Any:
        ...

    @overload
    def get(self, key: str, default: Any) -> None:
        ...

    def has(self, key: str) -> bool:
        ...


class DatabasePersistor:
    def __init__(self, db_: db.Database, task_id: Optional[db.TaskRowID]) -> None:
        self._db = db_
        self._task = task_id

    def get(self, key: str, default: Any = db.Undefined) -> Any:
        val = self._db.load_persist(key, self._task)
        if val is db.Undefined:
            val = default
        if val is db.Undefined:
            raise KeyError(f'No persisted value for key "{key}"')
        return val

    def set(self, key: str, value: Any) -> None:
        return self._db.set_persist(key, self._task, value)

    def has(self, key: str) -> bool:
        return db.Undefined is not self._db.load_persist(key, self._task)


_TASK_LOCAL_PERSIST = contextvars.ContextVar[IPersistentDataAccess]('_TASK_LOCAL_PERSIST')
_GLOBAL_PERSIST = contextvars.ContextVar[IPersistentDataAccess]('_GLOBAL_PERSIST')


@contextmanager
def set_task_local_persistor(p: IPersistentDataAccess) -> Iterator[IPersistentDataAccess]:
    t = _TASK_LOCAL_PERSIST.set(p)
    try:
        yield p
    finally:
        _TASK_LOCAL_PERSIST.reset(t)


@contextmanager
def set_global_persistor(p: IPersistentDataAccess) -> Iterator[IPersistentDataAccess]:
    t = _GLOBAL_PERSIST.set(p)
    try:
        yield p
    finally:
        _GLOBAL_PERSIST.reset(t)


def get_task_local_persistor() -> IPersistentDataAccess:
    return _TASK_LOCAL_PERSIST.get()


def get_global_persistor() -> IPersistentDataAccess:
    return _GLOBAL_PERSIST.get()


class _ImplicitPersistenceItem(NamedTuple):
    cv_: contextvars.ContextVar[IPersistentDataAccess]

    def get(self, key: str, default: Any = db.Undefined) -> Any:
        return self.cv_.get().get(key, default)

    def set(self, key: str, value: Any) -> None:
        self.cv_.get().set(key, value)

    def has(self, key: str) -> bool:
        return self.cv_.get().has(key)


if TYPE_CHECKING:
    typecheck(IPersistentDataAccess)(DatabasePersistor)
    typecheck(IPersistentDataAccess)(_ImplicitPersistenceItem)


class _PersistExt(BaseExtension[None, None, None]):
    dagon_ext_name = 'dagon.persist'
    dagon_ext_requires = ['dagon.db']

    @asynccontextmanager
    async def global_context(self, graph: OpaqueTaskGraphView) -> AsyncIterator[None]:
        gdb = db.global_context_data()
        assert gdb
        with set_global_persistor(DatabasePersistor(gdb.database, None)):
            yield

    @asynccontextmanager
    async def task_context(self, task: OpaqueTask) -> AsyncIterator[None]:
        gdb = db.global_context_data()
        tdb = db.task_context_data()
        assert tdb and gdb
        with set_task_local_persistor(DatabasePersistor(gdb.database, tdb.task_id)):
            yield


class _ImplicitPersistence(NamedTuple):
    globl: IPersistentDataAccess = _ImplicitPersistenceItem(_GLOBAL_PERSIST)
    local: IPersistentDataAccess = _ImplicitPersistenceItem(_TASK_LOCAL_PERSIST)


_imp_persist = _ImplicitPersistence()

unused(_PersistExt)

globl = _imp_persist.globl
local = _imp_persist.local



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dagon',
 'dagon.ar',
 'dagon.core',
 'dagon.db',
 'dagon.event',
 'dagon.ext',
 'dagon.fs',
 'dagon.inspect',
 'dagon.option',
 'dagon.persist',
 'dagon.pool',
 'dagon.proc',
 'dagon.script_mode',
 'dagon.storage',
 'dagon.task',
 'dagon.tool',
 'dagon.ui',
 'dagon.util']

package_data = \
{'': ['*']}

install_requires = \
['typing-extensions>=4.0.1,<5.0.0']

extras_require = \
{':python_version < "3.10"': ['importlib-metadata']}

entry_points = \
{'console_scripts': ['dagon = dagon.tool.main:start',
                     'dagon-inspect = dagon.inspect.main:start'],
 'dagon.extensions': ['dagon.db = dagon.db:_DatabaseExt',
                      'dagon.events = dagon.event:_EventsExt',
                      'dagon.options = dagon.option.ext:_OptionsExt',
                      'dagon.persist = dagon.persist:_PersistExt',
                      'dagon.pools = dagon.pool:_PoolsExt',
                      'dagon.ui = dagon.ui:_Ext'],
 'dagon.uis': ['dagon.ui.fancy = dagon.ui.fancy:FancyUI',
               'dagon.ui.simple = dagon.ui.simple:SimpleUI']}

setup_kwargs = {
    'name': 'dagon',
    'version': '0.9.3',
    'description': 'An asynchronous task graph execution system',
    'long_description': '# Dagon - An Asynchronous Task Graph Execution Engine\n\n*Dagon* is a job execution system designed for speed, flexibility,\nexpressiveness, and correctness with two overarching goals in mind:\n\n1. It should be *easy* to use *correctly*.\n2. It should be *hard* to use *incorrectly*.\n\nDagon itself can be used as a development tool, a CI tool, an application\nframework, or a library for creating declarative task-graph programs and\npipelines.\n\nFor more information, refer to the `docs/` subdirectory.\n',
    'author': 'vector-of-bool',
    'author_email': 'vectorofbool@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/vector-of-bool/dagon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import tarfile
import tempfile
import zipfile
from pathlib import Path

import pytest

from dagon import ar

PKG_ROOT = Path(__file__).absolute().parent.parent


@pytest.mark.asyncio
async def test_tar_create_expand() -> None:
    tdir = tempfile.mkdtemp()
    tar_path = await ar.create_from_dir(PKG_ROOT, destination=Path(tdir) / 'test.tgz')
    assert tar_path.is_file()
    print('Created tar at', tar_path)

    with tarfile.open(tar_path) as tfd:
        tl_py = tfd.getmember('ar/ar_test.py')
        assert tl_py.isfile()
        tl_py_fd = tfd.extractfile(tl_py)
        assert tl_py_fd
        exp_content = PKG_ROOT.joinpath('ar/ar_test.py').read_bytes()
        assert tl_py_fd.read() == exp_content

    exp_dir = Path(tempfile.mkdtemp())
    await ar.expand(tar_path, destination=exp_dir, if_exists='merge')
    assert (exp_dir / 'ar/ar_test.py').is_file()

    # Trying to expand again will raise
    with pytest.raises(FileExistsError):
        await ar.expand(tar_path, destination=exp_dir)

    # Merge mode will complain about overwriting files
    with pytest.raises(FileExistsError):
        await ar.expand(tar_path, destination=exp_dir, if_exists='merge')

    await ar.expand(tar_path, destination=exp_dir, if_exists='merge', if_file_exists='replace')

    await ar.expand(tar_path, destination=exp_dir, if_exists='merge', if_file_exists='keep')


@pytest.mark.asyncio
async def test_zip_create_expand() -> None:
    # Check that the default extension is .zip when the format is Zip
    zip_path = await ar.create_from_dir(PKG_ROOT, format='zip')
    assert zip_path.suffix == '.zip'
    await ar.fs.remove(zip_path)

    tdir = Path(tempfile.mkdtemp())
    zip_path = await ar.create_from_dir(PKG_ROOT, destination=tdir / 'test.zip', format='zip')

    assert zip_path.exists()

    with zipfile.ZipFile(zip_path) as zfd:
        tl_py = zfd.getinfo('ar/ar_test.py')
        content = (PKG_ROOT / 'ar/ar_test.py').read_bytes()
        assert zfd.read(tl_py) == content

    exp_dir = Path(tempfile.mkdtemp())
    await ar.expand(zip_path, destination=exp_dir, if_exists='merge')
    assert (exp_dir / 'ar/ar_test.py').is_file()

    # Trying to expand again will raise
    with pytest.raises(FileExistsError):
        await ar.expand(zip_path, destination=exp_dir)

    # Merge mode will complain about overwriting files
    with pytest.raises(FileExistsError):
        await ar.expand(zip_path, destination=exp_dir, if_exists='merge')

    await ar.expand(zip_path, destination=exp_dir, if_exists='merge', if_file_exists='replace')

    await ar.expand(zip_path, destination=exp_dir, if_exists='merge', if_file_exists='keep')
